# coding=utf-8
"""Base class for Ladybug Data Collections."""

from __future__ import division

from .header import Header
from .datatype.base import DataTypeBase
from .datatype import TYPESDICT, BASETYPES

try:
    from collections.abc import Iterable  # python < 3.7
except ImportError:
    from collections import Iterable  # python >= 3.8
from string import ascii_lowercase
import math

try:
    from itertools import izip as zip  # python 2
except ImportError:
    xrange = range  # python 3


class BaseCollection(object):
    """Base class for all Data Collections.

    Args:
        header: A Ladybug Header object.
        values: A list of values.
        datetimes: A list of Ladybug DateTime objects that aligns with
            the list of values.
    """

    __slots__ = ('_header', '_values', '_datetimes', '_validated_a_period')
    _collection_type = None
    _mutable = True
    _enumeration = None

    def __init__(self, header, values, datetimes):
        """Initialize base collection.
        """
        assert isinstance(header, Header), \
            'header must be a Ladybug Header object. Got {}'.format(type(header))
        assert isinstance(datetimes, Iterable) \
            and not isinstance(datetimes, (str, dict, bytes, bytearray)), \
            'datetimes should be a list or tuple. Got {}'.format(type(datetimes))

        self._header = header
        self._datetimes = tuple(datetimes)
        self.values = values
        self._validated_a_period = False

    @classmethod
    def from_dict(cls, data):
        """Create a Data Collection from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
                "header": {},  # Ladybug Header
                "values": [],  # array of values
                "datetimes": [],  # array of datetimes
                "validated_a_period": True  # boolean for valid analysis_period
            }
        """
        assert 'type' in data, 'Required key "type" is missing!'
        assert cls._collection_type == data['type'], \
            'Expected {}. Got {}.'.format(cls._collection_type, data['type'])
        assert 'header' in data, 'Required keyword "header" is missing!'
        assert 'values' in data, 'Required keyword "values" is missing!'
        assert 'datetimes' in data, 'Required keyword "datetimes" is missing!'
        coll = cls(Header.from_dict(data['header']), data['values'], data['datetimes'])
        if 'validated_a_period' in data:
            coll._validated_a_period = data['validated_a_period']
        return coll

    @property
    def header(self):
        """Get the header for this collection."""
        return self._header

    @property
    def datetimes(self):
        """Get a tuple of datetimes for this collection, which align with the values."""
        return self._datetimes

    @property
    def values(self):
        """Get a tuple of numerical values for this collection."""
        return tuple(self._values)

    @values.setter
    def values(self, values):
        self._check_values(values)
        self._values = list(values)

    @property
    def validated_a_period(self):
        """Boolean for whether the header analysis_period is validated against datetimes.

        This will always be True when a collection is derived from a continuous one.
        """
        return self._validated_a_period

    @property
    def datetime_strings(self):
        """Get a list of datetime strings for this collection.

        These provide a human-readable way to interpret the datetimes.
        """
        return [str(d) for d in self.datetimes]

    @property
    def bounds(self):
        """Get a tuple of two value as (min, max) of the data."""
        return (min(self._values), max(self._values))

    @property
    def min(self):
        """Get the min of the Data Collection values."""
        return min(self._values)

    @property
    def max(self):
        """Get the max of the Data Collection values."""
        return max(self._values)

    @property
    def average(self):
        """Get the average of the Data Collection values."""
        return sum(self._values) / len(self._values)

    @property
    def median(self):
        """Get the median of the Data Collection values."""
        return self._percentile(self._values, 50)

    @property
    def total(self):
        """Get the total of the Data Collection values."""
        return sum(self._values)

    def convert_to_unit(self, unit):
        """Convert the Data Collection to the input unit.

        Note that this mutates the data collection object, which can have unintended
        consequences depending on how the data collection is used. Use to_unit to
        get a new instance of a collection without mutating this one.
        """
        self._values = self._header.data_type.to_unit(
            self._values, unit, self._header.unit)
        self._header._unit = unit

    def convert_to_ip(self):
        """Convert the Data Collection to IP units.

        Note that this mutates the data collection object, which can have unintended
        consequences depending on how the data collection is used. Use to_ip to
        get a new instance of a collection without mutating this one.
        """
        self._values, self._header._unit = self._header.data_type.to_ip(
            self._values, self._header.unit)

    def convert_to_si(self):
        """Convert the Data Collection to SI units.

        Note that this mutates the data collection object, which can have unintended
        consequences depending on how the data collection is used. Use to_si to
        get a new instance of a collection without mutating this one.
        """
        self._values, self._header._unit = self._header.data_type.to_si(
            self._values, self._header.unit)

    def to_unit(self, unit):
        """Get a Data Collection in the input unit.

        Args:
            unit: Text for the unit to convert the data to (eg. 'C' or 'kWh'). This
                unit must appear under the data collection's header.data_type.units.
        """
        new_data_c = self.duplicate()
        new_data_c.convert_to_unit(unit)
        return new_data_c

    def to_ip(self):
        """Get a Data Collection in IP units."""
        new_data_c = self.duplicate()
        new_data_c.convert_to_ip()
        return new_data_c

    def to_si(self):
        """Get a Data Collection in SI units."""
        new_data_c = self.duplicate()
        new_data_c.convert_to_si()
        return new_data_c

    def is_in_data_type_range(self, raise_exception=True):
        """Check if collection values are in the range for the data_type.

        If this method returns False, the collection's values are physically or
        mathematically impossible for the data_type (eg. temperature below
        absolute zero).

        Args:
            raise_exception: Boolean to note whether an exception should be raised
                if an impossible value is found. (Default: True).
        """
        return self._header.data_type.is_in_range(
            self._values, self._header.unit, raise_exception)

    def to_mutable(self):
        """Get a mutable version of this collection."""
        return self.duplicate()

    def to_immutable(self):
        """Get an immutable version of this collection."""
        if self._enumeration is None:
            self._get_mutable_enumeration()
        col_obj = self._enumeration['immutable'][self._collection_type]
        new_obj = col_obj(self.header, self.values, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def normalize_by_area(self, area, area_unit):
        """Get a Data Collection that is normalized by an area value.

        Note that this method will raise a ValueError if the data type in the header
        of the data collection does not have a normalized_type. Also note that a
        ZeroDivisionError will be raised if the input area is equal to 0.

        Args:
            area: Number representing area by which all of the data is normalized.
            area_unit: Text for the units that the area value is in. Acceptable
                inputs include 'm2', 'ft2' and any other unit that is supported
                in the normalized_type of this datacollection's data type.
        """
        # get an instance of the normalized data type
        head = self.header
        norm_type_class = head.data_type.normalized_type
        assert norm_type_class is not None, \
            'Data type "{}" cannot be normalized by area to yield a useful '\
            'metric.'.format(head.data_type)

        # create the new data collection and assign normalized values
        new_data_c = self.duplicate()
        new_data_c._values = [val / area for val in self.values]

        # normalize the data type and unit in the header
        new_data_c._header._unit = '{}/{}'.format(head.unit, area_unit) \
            if '/' not in head.unit else '{}-{}'.format(head.unit, area_unit)
        new_data_c._header._data_type = norm_type_class()
        new_data_c._header._data_type.is_unit_acceptable(new_data_c._header._unit)
        if 'type' in head.metadata:  # key used to identify sophisticated data types
            new_data_c._header.metadata['type'] = \
                '{} {}'.format(new_data_c._header.metadata['type'], 'Intensity')
        return new_data_c

    def aggregate_by_area(self, area, area_unit):
        """Get a Data Collection that is aggregated by an area value.

        Note that this method will raise a ValueError if the data type in the header
        of the data collection is not a normalized_type of another data type.

        Args:
            area: Number representing area by which all of the data is aggregated.
            area_unit: Text for the units that the area value is in. Acceptable
                inputs include 'm2', 'ft2' and any other unit that is supported
                in the normalized_type of destination datacollection's data type.
        """
        # get an instance of the time aggregated data type
        head = self.header
        dat_type, agg_class = head.data_type, None
        # first see if there's a specific data type for the current one
        for typ_clss in TYPESDICT.values():
            if typ_clss._normalized_type is None:
                continue
            elif dat_type.__class__ == typ_clss._normalized_type:
                agg_class = typ_clss
                break
        # then, check to see if there's any base type
        if agg_class is None:
            for typ_clss_name in BASETYPES:
                typ_clss = TYPESDICT[typ_clss_name]
                if typ_clss._normalized_type is None:
                    continue
                elif isinstance(dat_type, typ_clss._normalized_type):
                    agg_class = typ_clss
                    break
        # if nothing was found, throw an error
        if agg_class is None:
            raise ValueError('Data type "{}" does not have a area-'
                             'aggregated metric.'.format(head.data_type))

        # create the new data collection and assign normalized values
        new_data_c = self.duplicate()
        new_data_c._values = [val * area for val in self.values]

        # normalize the data type and unit in the header
        new_data_c._header._unit = head.unit.replace('/{}'.format(area_unit), '') \
            if '-' not in head.unit.split('/')[-1] else \
            head.unit.replace('-{}'.format(area_unit), '')
        new_data_c._header._data_type = agg_class()
        new_data_c._header._data_type.is_unit_acceptable(new_data_c._header._unit)
        if 'type' in head.metadata:  # key used to identify sophisticated data types
            new_data_c._header.metadata['type'] = \
                new_data_c._header.metadata['type'].replace(' Intensity', '')
        return new_data_c

    def highest_values(self, count):
        """Get a list of the the highest values of the Data Collection and their indices.

        This is useful for situations where one needs to know the times of
        the year when the largest values of a data collection occur.  For example,
        there is a European daylight code that requires an analysis for the hours
        of the year with the greatest exterior illuminance level.  This method
        can be used to help build a schedule for such a study.

        Args:
            count: Integer representing the number of highest values to account for.

        Returns:
            A tuple with two elements.

            -   highest_values:
                The n highest values in data list, ordered from
                highest to lowest.

            -   highest_values_index:
                Indices of the n highest values in data
                list, ordered from highest to lowest.
        """
        count = int(count)
        assert count <= len(self._values), \
            'count must be smaller than or equal to values length. {} > {}.'.format(
                count, len(self._values))
        assert count > 0, \
            'count must be greater than 0. Got {}.'.format(count)
        highest_values = sorted(self._values, reverse=True)[0:count]
        highest_values_index = sorted(list(xrange(len(self._values))),
                                      key=lambda k: self._values[k],
                                      reverse=True)[0:count]
        return highest_values, highest_values_index

    def lowest_values(self, count):
        """Get a list of the the lowest values of the Data Collection and their indices.

        This is useful for situations where one needs to know the times of
        the year when the smallest values of a data collection occur.

        Args:
            count: Integer representing the number of lowest values to account for.

        Returns:
            A tuple with two elements.

            -   highest_values:
                The n lowest values in data list, ordered from
                lowest to lowest.
            -   lowest_values_index:
                Indices of the n lowest values in data
                list, ordered from lowest to lowest.
        """
        count = int(count)
        assert count <= len(self._values), \
            'count must be <= to Data Collection len. {} > {}.'.format(
                count, len(self._values))
        assert count > 0, \
            'count must be greater than 0. Got {}.'.format(count)
        lowest_values = sorted(self._values)[0:count]
        lowest_values_index = sorted(list(xrange(len(self._values))),
                                     key=lambda k: self._values[k])[0:count]
        return lowest_values, lowest_values_index

    def percentile(self, percentile):
        """Get a value representing a the input percentile of the Data Collection.

        Args:
            percentile: A float value from 0 to 100 representing the
                requested percentile.

        Returns:
            The Data Collection value at the input percentile
        """
        assert 0 <= percentile <= 100, \
            'percentile must be between 0 and 100. Got {}'.format(percentile)
        return self._percentile(self._values, percentile)

    def filter_by_conditional_statement(self, statement):
        """Filter the Data Collection based on a conditional statement.

        Args:
            statement: A conditional statement as a string (e.g. a > 25 and a%5 == 0).
                The variable should always be named as 'a' (without quotations).

        Returns:
            A new Data Collection containing only the filtered data.
        """
        _filt_values, _filt_datetimes = self._filter_by_statement(statement)
        if self._enumeration is None:
            self._get_mutable_enumeration()
        col_obj = self._enumeration['mutable'][self._collection_type]
        try:
            collection = col_obj(self.header.duplicate(), _filt_values, _filt_datetimes)
        except AssertionError as e:
            raise AssertionError('No value meets the conditional statement.'
                                 '\n{}'.format(e))
        collection._validated_a_period = self._validated_a_period
        return collection

    def filter_by_range(self, greater_than=float('-inf'), less_than=float('inf')):
        """Filter the Data Collection based on whether values fall within a given range.

        This is similar to the filter_by_conditional_statement but is often much
        faster since it does not have all of the flexibility of the conditional
        statement and uses native Python operators instead of eval() statements.

        Args:
            greater_than: A number which the data collection values should be
                greater than in order to be included in the output
                collection. (Default: Negative Infinity).
            less_than: A number which the data collection values should be less than
                in order to be included in the output collection. (Default: Infinity).

        Returns:
            A new Data Collection with filtered data.
        """
        _filt_values, _filt_datetimes = self._filter_by_range(greater_than, less_than)
        if self._enumeration is None:
            self._get_mutable_enumeration()
        col_obj = self._enumeration['mutable'][self._collection_type]
        try:
            collection = col_obj(self.header.duplicate(), _filt_values, _filt_datetimes)
        except AssertionError as e:
            raise AssertionError(
                'No value was found in the range {} to {}.\n{}'.format(
                    greater_than, less_than, e))
        collection._validated_a_period = self._validated_a_period
        return collection

    def filter_by_pattern(self, pattern):
        """Filter the Data Collection based on a list of booleans.

        Args:
            pattern: A list of True/False values. Typically, this is a list
                with a length matching the length of the Data Collections values
                but it can also be a pattern to be repeated over the Data Collection.

        Returns:
            A new Data Collection with filtered data.
        """
        _filt_values, _filt_datetimes = self._filter_by_pattern(pattern)
        if self._enumeration is None:
            self._get_mutable_enumeration()
        col_obj = self._enumeration['mutable'][self._collection_type]
        collection = col_obj(self.header.duplicate(), _filt_values, _filt_datetimes)
        collection._validated_a_period = self._validated_a_period
        return collection

    def is_collection_aligned(self, data_collection):
        """Check if this Data Collection is aligned with another.

        Aligned Data Collections are of the same Data Collection class, have the
        same number of values and have matching datetimes.

        Args:
            data_collection: The Data Collection for which alignment will be tested.

        Returns:
            True if collections are aligned, False if not aligned
        """
        if self._collection_type != data_collection._collection_type:
            return False
        elif len(self.values) != len(data_collection.values):
            return False
        elif self.datetimes != data_collection.datetimes:
            return False
        else:
            return True

    def get_aligned_collection(self, value=0, data_type=None, unit=None, mutable=None):
        """Get a collection aligned with this one composed of one repeated value.

        Aligned Data Collections are of the same Data Collection class, have the same
        number of values and have matching datetimes.

        Args:
            value: A value to be repeated in the aligned collection values or
                A list of values that has the same length as this collection.
                Default: 0.
            data_type: The data type of the aligned collection. Default is to
                use the data type of this collection.
            unit: The unit of the aligned collection. Default is to
                use the unit of this collection or the base unit of the
                input data_type (if it exists).
            mutable: An optional Boolean to set whether the returned aligned
                collection is mutable (True) or immutable (False). The default is
                None, which will simply set the aligned collection to have the
                same mutability as the starting collection.
        """
        # set up the header of the new collection
        header = self._check_aligned_header(data_type, unit)

        # set up the values of the new collection
        values = self._check_aligned_value(value)

        # get the correct base class for the aligned collection (mutable or immutable)
        if mutable is None:
            collection = self.__class__(header, values, self.datetimes)
        else:
            if self._enumeration is None:
                self._get_mutable_enumeration()
            if not mutable:
                col_obj = self._enumeration['immutable'][self._collection_type]
            else:
                col_obj = self._enumeration['mutable'][self._collection_type]
            collection = col_obj(header, values, self.datetimes)
        collection._validated_a_period = self._validated_a_period
        return collection

    def is_metadata_aligned(self, data_collection):
        """Check if the metadata in this Data Collection header is aligned with another.

        Aligned metadata means that the number of metadata items is the same
        between the two collections.

        Args:
            data_collection: The Data Collection for which metadata alignment
                will be tested.

        Returns:
            True if the metadata in the collections are aligned, False if not aligned.
        """
        return len(self.header.metadata) == len(data_collection.header.metadata)

    def duplicate(self):
        """Get a copy of this Data Collection."""
        return self.__copy__()

    def to_dict(self):
        """Convert Data Collection to a dictionary."""
        return {
            'header': self.header.to_dict(),
            'values': self._values,
            'datetimes': self.datetimes,
            'validated_a_period': self._validated_a_period,
            'type': self._collection_type
        }

    @staticmethod
    def filter_collections_by_statement(data_collections, statement):
        """Generate a filtered data collections according to a conditional statement.

        Args:
            data_collections: A list of aligned Data Collections to be evaluated
                against the statement.
            statement: A conditional statement as a string (e.g. a>25 and a%5==0).
                The variable should always be named as 'a' (without quotations).

        Returns:
            collections -- A list of Data Collections that have been filtered based
            on the statement.
        """
        pattern = BaseCollection.pattern_from_collections_and_statement(
            data_collections, statement)
        try:
            collections = [coll.filter_by_pattern(pattern) for coll in data_collections]
        except AssertionError as e:
            raise AssertionError('No value meets the conditional statement.'
                                 '\n{}'.format(e))
        return collections

    @staticmethod
    def pattern_from_collections_and_statement(data_collections, statement):
        """Generate a list of booleans from data collections and a conditional statement.

        Args:
            data_collections: A list of aligned Data Collections to be evaluated
                against the statement.
            statement: A conditional statement as a string (e.g. a>25 and a%5==0).
                The variable should always be named as 'a' (without quotations).

        Returns:
            pattern -- A list of True/False booleans with the length of the
            Data Collections where True meets the conditional statement
            and False does not.
        """
        BaseCollection.are_collections_aligned(data_collections)
        correct_var = BaseCollection._check_conditional_statement(
            statement, len(data_collections))

        # replace the operators of the statement with non-alphanumeric characters
        # necessary to avoid replacing the characters of the operators
        num_statement_clean = BaseCollection._replace_operators(statement)

        pattern = []
        for i in xrange(len(data_collections[0])):
            num_statement = num_statement_clean
            # replace the variable names with their numerical values
            for j, coll in enumerate(data_collections):
                var = correct_var[j]
                num_statement = num_statement.replace(var, str(coll[i]))
            # put back the operators
            num_statement = BaseCollection._restore_operators(num_statement)
            pattern.append(eval(num_statement, {}))
        return pattern

    @staticmethod
    def are_collections_aligned(data_collections, raise_exception=True):
        """Test if a series of Data Collections are aligned with one another.

        Aligned Data Collections are of the same Data Collection class, have the
        same number of values and have matching datetimes.

        Args:
            data_collections: A list of Data Collections for which alignment
                will be tested.
            raise_exception: Boolean to note if an exception should be raised
                when collections are not aligned with one another.

        Returns:
            True if collections are aligned, False if not aligned
        """
        if len(data_collections) > 1:
            first_coll = data_collections[0]
            for coll in data_collections[1:]:
                if not first_coll.is_collection_aligned(coll):
                    if raise_exception:
                        error_msg = '{} Data Collection is not aligned with '\
                            '{} Data Collection.'.format(
                                first_coll.header.data_type, coll.header.data_type)
                        raise ValueError(error_msg)
                    return False
        return True

    @staticmethod
    def compute_function_aligned(funct, data_collections, data_type, unit):
        """Compute a function with a list of aligned data collections or values.

        Args:
            funct: A function with a single numerical value as output and one or
                more numerical values as input.
            data_collections: A list with a length equal to the number of arguments
                for the function. Items of the list can be either Data Collections
                or individual values to be used at each datetime of other collections.
            data_type: An instance of a Ladybug data type that describes the results
                of the funct.
            unit: The units of the funct results.

        Returns:
            A Data Collection with the results function. If all items in this list of
            data_collections are individual values, only a single value will be returned.

        Usage:

        .. code-block:: python

            from ladybug.datacollection import HourlyContinuousCollection
            from ladybug.epw import EPW
            from ladybug.psychrometrics import humid_ratio_from_db_rh
            from ladybug.datatype.percentage import HumidityRatio

            epw_file_path = './epws/denver.epw'
            denver_epw = EPW(epw_file_path)
            pressure_at_denver = 85000
            hr_inputs = [denver_epw.dry_bulb_temperature,
                         denver_epw.relative_humidity,
                         pressure_at_denver]
            humid_ratio = HourlyContinuousCollection.compute_function_aligned(
                humid_ratio_from_db_rh, hr_inputs, HumidityRatio(), 'fraction')
            # humid_ratio will be a Data Collection of humidity ratios at Denver
        """
        # check that all inputs are either data collections or floats
        data_colls = []
        for i, func_input in enumerate(data_collections):
            if isinstance(func_input, BaseCollection):
                data_colls.append(func_input)
            else:
                try:
                    data_collections[i] = float(func_input)
                except ValueError:
                    raise TypeError('Expected a number or a Data Collection. '
                                    'Got {}'.format(type(func_input)))

        # run the function and return the result
        if len(data_colls) == 0:
            return funct(*data_collections)
        else:
            BaseCollection.are_collections_aligned(data_colls)
            val_len = len(data_colls[0].values)
            for i, col in enumerate(data_collections):
                data_collections[i] = [col] * val_len if isinstance(col, float) else col
            result = data_colls[0].get_aligned_collection(data_type=data_type, unit=unit)
            if not result._mutable:
                result = result.to_mutable()
            for i in xrange(val_len):
                result[i] = funct(*[col[i] for col in data_collections])
            return result

    @staticmethod
    def are_metadatas_aligned(data_collections, raise_exception=True):
        """Test if a series of Data Collections have aligned metadata.

        Aligned metadata means that the number of metadata items is the same
        between the two collections.

        Args:
            data_collections: A list of Data Collections for which metadata
                alignment will be tested.
            raise_exception: Boolean to note if an exception should be raised
                when collection metadatas are not aligned with one another.

        Returns:
            True if metadatas are aligned, False if not aligned
        """
        if len(data_collections) > 1:
            first_coll = data_collections[0]
            for coll in data_collections[1:]:
                if not first_coll.is_metadata_aligned(coll):
                    if raise_exception:
                        error_msg = '{} Data Collection metadata is not aligned with '\
                            '{} Data Collection.'.format(
                                first_coll.header.data_type, coll.header.data_type)
                        raise ValueError(error_msg)
                    return False
        return True

    def _time_aggregated_collection(self, timestep):
        """Get a time-aggregated version of this collection."""
        # get an instance of the time aggregated data type
        head = self.header
        time_class = head.data_type.time_aggregated_type
        assert time_class is not None, \
            'Data type "{}" cannot be time aggregated to yield a useful '\
            'metric.'.format(head.data_type)

        # create the new data collection and assign normalized values
        new_data_c = self.to_unit(head.data_type.units[0])
        factor = head.data_type.time_aggregated_factor / timestep
        new_data_c._values = [val * factor for val in new_data_c._values]
        new_data_c._header._data_type = time_class()
        new_data_c._header._unit = new_data_c._header._data_type.units[0]
        return new_data_c

    def _time_rate_of_change_collection(self, timestep):
        """Get a time-rate-of-change version of this collection."""
        # get an instance of the time aggregated data type
        head = self.header
        dat_type, time_class = head.data_type, None
        # see if there's a specific data type for the current one
        for typ_clss_name in BASETYPES:
            typ_clss = TYPESDICT[typ_clss_name]
            if typ_clss._time_aggregated_type is None:
                continue
            elif isinstance(dat_type, typ_clss._time_aggregated_type):
                time_class = typ_clss
                break
        # if nothing was found, throw an error
        if time_class is None:
            raise ValueError('Data type "{}" does not have a time-rate-of-'
                             'change metric.'.format(head.data_type))

        # create the new data collection and assign normalized values
        new_data_c = self.to_unit(head.data_type.units[0])
        factor = typ_clss._time_aggregated_factor / timestep
        new_data_c._values = [val / factor for val in new_data_c._values]
        new_data_c._header._data_type = time_class()
        new_data_c._header._unit = new_data_c._header._data_type.units[0]
        return new_data_c

    @staticmethod
    def _check_conditional_statement(statement, num_collections):
        """Method to check conditional statements to be sure that they are valid.

        Args:
            statement: A conditional statement as a string (e.g. a>25 and a%5==0).
                The variable should always be named as 'a' (without quotations).
            num_collections: An integer representing the number of data collections
                that the statement will be evaluating.

        Returns:
            correct_var -- A list of the correct variable names that should be
                used within the statement (eg. ['a', 'b', 'c'])
        """
        # Determine what the list of variables should be based on the num_collections
        correct_var = list(ascii_lowercase)[:num_collections]

        # Clean out the operators of the statement
        st_statement = BaseCollection._remove_operators(statement)
        parsed_st = [s for s in st_statement if s.isalpha()]

        # Perform the check
        for var in parsed_st:
            if var not in correct_var:
                raise ValueError(
                    'Invalid conditional statement: {}\n '
                    'Statement should be a valid Python statement'
                    ' and the variables should be named as follows: {}'.format(
                        statement, ', '.join(correct_var))
                )
        return correct_var

    @staticmethod
    def _remove_operators(statement):
        """Remove logical operators from a statement."""
        return statement.lower().replace("and", "").replace("or", "") \
            .replace("not", "").replace("in", "").replace("is", "")

    @staticmethod
    def _replace_operators(statement):
        """Replace logical operators of a statement with non-alphanumeric characters."""
        return statement.lower().replace("and", "&&").replace("or", "||") \
            .replace("not", "~").replace("in", "<<").replace("is", "$")

    @staticmethod
    def _restore_operators(statement):
        """Restore python logical operators from previously replaced ones."""
        return statement.replace("&&", "and").replace("||", "or") \
            .replace("~", "not").replace("<<", "in").replace("$", "is")

    @staticmethod
    def linspace(start, stop, num):
        """Get evenly spaced numbers calculated over the interval start, stop.

        This method is similar to native Python range except that it takes a number of
        divisions instead of a step. It is also equivalent to numpy's linspace method.

        Args:
            start: Start interval index as integer or float.
            stop: Stop interval index as integer or float.
            num: Number of divisions as integer.

        Returns:
            A list of numbers.

        Usage:

        .. code-block:: python

            from BaseCollection import linspace

            linspace(0, 5, 6)
            # >> [0., 1., 2., 3., 4., 5.]
        """
        try:
            delta = stop - start
            return [i * (delta / (num - 1)) + start for i in range(num)]
        except ZeroDivisionError:
            return [start]

    @staticmethod
    def arange(start, stop, step):
        """Return evenly spaced fractional or whole values within a given interval.

        This function acts like the Python range method, but can also account for
        fractional values. It is equivalent to the numpy.arange function.

        Args:
            start: Number for inclusive start of interval.
            stop: Number for exclusive end of interval.
            step: Number for step size of interval.

        Returns:
            Generator of evenly spaced values.

        Usage:

        .. code-block:: python

            from BaseCollection import arange

            arange(1, 351, 50)
            # >> [1, 51, 101, 151, 201, 251, 301]
        """
        val = start

        if start <= stop:
            def ineq(a, b):
                return a < b
        else:
            def ineq(a, b):
                return a > b

        while ineq(val, stop) and abs(val - stop) > 1e-10:
            yield val
            val += step

    @staticmethod
    def histogram(values, bins, key=None):
        """Compute the frequency histogram from a list of values.

        The data is binned inclusive of the lower bound but exclusive of the upper bound
        for intervals. See usage for example of losing the last number in the following
        dataset because of exclusive upper bound.

        Args:
            values: Set of numerical data as a list.
            bins: A monotonically increasing array of uniform-width bin edges, excluding
                the rightmost edge.
            key: Optional parameter to define key to bin values by, as a function. If not
                provided, the histogram will be binned by the value.

        Returns:
            A list of lists representing the ordered values binned by frequency.

        Usage:

        .. code-block:: python

            from BaseCollection import histogram

            # Simple example
            histogram([0, 0, 0.9, 1, 1.5, 1.99, 2, 3], (0, 1, 2, 3))
            # >> [[0, 0, 0.9], [1, 1.5, 1.99], [2]]

            # With key parameter
            histogram(
                zip([0, 0, 0.9, 1, 1.5, 1.99],
                    ['a', 'b', 'c', 'd', 'e', 'f']),
                    (0, 1, 2), key=lambda k: k[0])
            # >> [[], [(0, a), (0, b), (0.9, c)], [(1, d), (1.5, e), (1.99, f)], []]
        """
        if key is None:
            def key(v):
                return v

        vals = sorted(values, key=key)
        min_bound, max_bound = min(bins), max(bins)
        bin_bound_num = len(bins)

        # Init histogram bins
        hist = [[] for i in range(bin_bound_num + 1)]
        bin_index = 0

        for val in vals:
            k = key(val)
            # Ignore values out of range
            if k < min_bound:
                hist[0].append(val)
                continue
            elif k >= max_bound:
                hist[-1].append(val)
                continue

            # This loop will iterate through the bin upper bounds.
            # If the value is within the bounds, the lower bound
            # of the bin_index is updated, and the loop is broken
            for i in range(bin_index, bin_bound_num - 1):
                if k < bins[i + 1]:
                    hist[i + 1].append(val)
                    bin_index = i
                    break
        return hist

    @staticmethod
    def histogram_circular(values, bins, hist_range=None, key=None):
        """Compute the frequency histogram from a list of circular values.

        Circular values refers to a set of values where there is no distinction between
        values at the lower or upper end of the range, for example angles in a circle, or
        time. The data is binned inclusive of the lower bound but exclusive of the upper
        bound for intervals.

        Args:
            values: Set of numerical data as a list.
            bins: An array of uniform-width bin edges, excluding the rightmost edge.
                These values do not have to be monotonically increasing.
            hist_range: Optional parameter to define the lower and upper range of the
                histogram as a tuple of numbers. If not provided the range is
                ``(min(key(values)), max(key(values))+1)``.
            key: Optional parameter to define key to bin values by, as a function. If not
                provided, the histogram will be binned by the value.

        Returns:
            A list of lists representing the ordered values binned by frequency.

        Usage:

        .. code-block:: python

            from BaseCollection import histogram_circular

            histogram_circular([358, 359, 0, 1, 2, 3], (358, 0, 3))
            # >> [[358, 359], [0, 1, 2]]
        """
        if key is None:
            def key(v):
                return v

        vals = sorted(values, key=key)

        if hist_range is None:
            hist_range = (key(vals[0]), key(vals[-1]) + 1)

        bin_bound_num = len(bins) - 1

        # Init histogram bins
        hist = [[] for i in range(bin_bound_num)]
        for val in vals:
            k = key(val)

            # Ignore values out of range
            if k < hist_range[0] or k >= hist_range[1]:
                continue

            # This loop will iterate through the bin upper bounds.
            # If the value is within the bounds, the loop is broken.
            # Since values at the end of the list can still be binned
            # into the earlier histogram bars for circular
            # data, we don't update the bin_index.
            for i in range(bin_bound_num):
                if bins[i] < bins[i + 1]:
                    if k >= bins[i] and k < bins[i + 1]:
                        hist[i].append(val)
                        break
                else:
                    # If the interval starts data from the end of the list,
                    # split the conditional checks into two to check two
                    # intervals.
                    interval1 = (k <= hist_range[1] and k >= bins[i])
                    interval2 = (k < bins[i + 1] and k >= hist_range[0])
                    if interval1 or interval2:
                        hist[i].append(val)
                        break

        return hist

    def _filter_by_statement(self, statement):
        """Filter the data collection based on a conditional statement."""
        self.__class__._check_conditional_statement(statement, 1)
        _filt_values, _filt_datetimes = [], []
        for i, a in enumerate(self._values):
            if eval(statement, {'a': a}):
                _filt_values.append(a)
                _filt_datetimes.append(self.datetimes[i])
        return _filt_values, _filt_datetimes

    def _filter_by_range(self, greater_than, less_than):
        """Filter the data collection based on a range."""
        _filt_values, _filt_datetimes = [], []
        for i, a in enumerate(self._values):
            if greater_than < a < less_than:
                _filt_values.append(a)
                _filt_datetimes.append(self.datetimes[i])
        return _filt_values, _filt_datetimes

    def _filter_by_pattern(self, pattern):
        """Filter the Filter the Data Collection based on a list of booleans."""
        try:
            _len = len(pattern)
        except TypeError:
            raise TypeError("pattern is not a list of Booleans. Got {}".format(
                type(pattern)))
        _filt_values = [d for i, d in enumerate(self._values) if pattern[i % _len]]
        _filt_datetimes = [d for i, d in enumerate(self.datetimes) if pattern[i % _len]]
        return _filt_values, _filt_datetimes

    def _check_values(self, values):
        """Check values whenever they come through the values setter."""
        assert isinstance(values, Iterable) and not \
            isinstance(values, (str, dict, bytes, bytearray)), \
            'values should be a list or tuple. Got {}'.format(type(values))
        assert len(values) == len(self.datetimes), \
            'Length of values list must match length of datetimes list. ' \
            '{} != {}'.format(len(values), len(self.datetimes))
        assert len(values) > 0, 'Data Collection must include at least one value'

    def _check_aligned_header(self, data_type, unit):
        """Check the header inputs whenever get_aligned_collection is called."""
        if data_type is not None:
            assert isinstance(data_type, DataTypeBase), \
                'data_type must be a Ladybug DataType. Got {}'.format(type(data_type))
            if unit is None:
                unit = data_type.units[0]
        else:
            data_type = self.header.data_type
            unit = unit or self.header.unit
        return Header(data_type, unit, self.header.analysis_period, self.header.metadata)

    def _check_aligned_value(self, value):
        """Check the value input whenever get_aligned_collection is called."""
        if isinstance(value, Iterable) and not isinstance(
                value, (str, dict, bytes, bytearray)):
            assert len(value) == len(self._values), "Length of value ({}) must match "\
                "the length of this collection's values ({})".format(
                    len(value), len(self._values))
            values = value
        else:
            values = [value] * len(self._values)
        return values

    def _percentile(self, values, percent, key=lambda x: x):
        """Find the percentile of a list of values.

        Args:
            values: A list of values for which percentiles are desired
            percent: A float value from 0 to 100 representing the requested percentile.
            key: optional key function to compute value from each element of N.

        Returns:
            The percentile of the values
        """
        vals = sorted(values)
        k = (len(vals) - 1) * (percent / 100)
        f = math.floor(k)
        c = math.ceil(k)
        if f == c:
            return key(vals[int(k)])
        d0 = key(vals[int(f)]) * (c - k)
        d1 = key(vals[int(c)]) * (k - f)
        return d0 + d1

    def _average(self, vals):
        return sum(vals) / len(vals)

    def _total(self, vals):
        return sum(vals)

    def _get_percentile_function(self, percentile):
        def percentile_function(vals):
            return self._percentile(vals, percentile)
        return percentile_function

    def _get_mutable_enumeration(self):
        self._enumeration = {'mutable': {}, 'immutable': {}}
        for clss in self._all_subclasses(BaseCollection):
            if clss._mutable:
                self._enumeration['mutable'][clss._collection_type] = clss
            else:
                self._enumeration['immutable'][clss._collection_type] = clss

    def _all_subclasses(self, clss):
        return set(clss.__subclasses__()).union(
            [s for c in clss.__subclasses__() for s in self._all_subclasses(c)])

    def __len__(self):
        return len(self._values)

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        self._values[key] = value

    def __iter__(self):
        return iter(self._values)

    def __contains__(self, item):
        return item in self._values

    def __add__(self, other):
        new_vals = self._add_values(other)
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def __sub__(self, other):
        new_vals = self._sub_values(other)
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def __mul__(self, other):
        new_vals = self._mul_values(other)
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def __div__(self, other):
        new_vals = self._div_values(other)
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def __truediv__(self, other):
        new_vals = self._div_values(other)
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def __neg__(self):
        new_vals = [-v_1 for v_1 in self._values]
        new_obj = self.__class__(self.header.duplicate(), new_vals, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def _add_values(self, other):
        if isinstance(other, (int, float)):
            new_vals = [v_1 + other for v_1 in self._values]
        else:
            assert self._collection_type == other._collection_type, \
                '{} cannot be added to {}'.format(self.__class__, other.__class__)
            assert len(self) == len(other), 'Length of DataCollections must match in ' \
                'order to add them together. {} != {}'.format(len(self), len(other))
            new_vals = [v_1 + v_2 for v_1, v_2 in zip(self._values, other._values)]
        return new_vals

    def _sub_values(self, other):
        if isinstance(other, (int, float)):
            new_vals = [v_1 - other for v_1 in self._values]
        else:
            assert self._collection_type == other._collection_type, \
                '{} cannot be subtracted from {}'.format(other.__class__, self.__class__)
            assert len(self) == len(other), 'Length of DataCollections must match ' \
                'to subtract one from the other. {} != {}'.format(len(self), len(other))
            new_vals = [v_1 - v_2 for v_1, v_2 in zip(self._values, other._values)]
        return new_vals

    def _mul_values(self, other):
        if isinstance(other, (int, float)):
            new_vals = [v_1 * other for v_1 in self._values]
        else:
            assert self._collection_type == other._collection_type, \
                '{} cannot be multiplied by {}'.format(other.__class__, self.__class__)
            assert len(self) == len(other), 'Length of DataCollections must match ' \
                'to multiply them together. {} != {}'.format(len(self), len(other))
            new_vals = [v_1 * v_2 for v_1, v_2 in zip(self._values, other._values)]
        return new_vals

    def _div_values(self, other):
        if isinstance(other, (int, float)):
            new_vals = [v_1 / other for v_1 in self._values]
        else:
            assert self._collection_type == other._collection_type, \
                '{} cannot be divided by {}'.format(other.__class__, self.__class__)
            assert len(self) == len(other), 'Length of DataCollections must match ' \
                'to divide them. {} != {}'.format(len(self), len(other))
            new_vals = [v_1 / v_2 for v_1, v_2 in zip(self._values, other._values)]
        return new_vals

    @property
    def is_continuous(self):
        """Boolean denoting whether the data collection is continuous."""
        return False

    @property
    def is_mutable(self):
        """Boolean denoting whether the data collection is mutable."""
        return self._mutable

    def __key(self):
        return self.header, self.values, self.datetimes, self.validated_a_period

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__key() == other.__key()

    def __ne__(self, value):
        return not self.__eq__(value)

    def __copy__(self):
        collection = self.__class__(
            self.header.duplicate(), list(self._values), self.datetimes)
        collection._validated_a_period = self._validated_a_period
        return collection

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """Discontinuous Collection representation."""
        return "Discontinuous Data Collection\n{} ({})\n...{} values...".format(
            self.header.data_type, self.header.unit, len(self._values))



# coding=utf-8
"""Speed data type."""
from __future__ import division

from .base import DataTypeBase
from .distance import Distance


class Speed(DataTypeBase):
    """Speed
    """
    _units = ('m/s', 'mph', 'km/h', 'knot', 'ft/s', 'ft/min')
    _si_units = ('m/s', 'km/h')
    _ip_units = ('mph', 'ft/s', 'ft/min')
    _min = 0
    _abbreviation = 'v'
    _time_aggregated_type = Distance
    _time_aggregated_factor = 3600

    def _m_s_to_mph(self, value):
        return value * 2.23694

    def _m_s_to_km_h(self, value):
        return value * 3.6

    def _m_s_to_knot(self, value):
        return value * 1.94384

    def _m_s_to_ft_s(self, value):
        return value * 3.28084

    def _m_s_to_ft_min(self, value):
        return value * 196.85

    def _mph_to_m_s(self, value):
        return value / 2.23694

    def _km_h_to_m_s(self, value):
        return value / 3.6

    def _knot_to_m_s(self, value):
        return value / 1.94384

    def _ft_s_to_m_s(self, value):
        return value / 3.28084

    def _ft_min_to_m_s(self, value):
        return value / 196.85

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('m/s', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'mph', from_unit), 'mph'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'm/s', from_unit), 'm/s'


class WindSpeed(Speed):
    _abbreviation = 'WS'


class AirSpeed(Speed):
    _abbreviation = 'vair'



# coding=utf-8
"""Mass data type."""
from __future__ import division

from .base import DataTypeBase


class Mass(DataTypeBase):
    """Mass
    """
    _units = ('kg', 'lb', 'g', 'tonne', 'ton', 'oz')
    _si_units = ('kg', 'g', 'tonne')
    _ip_units = ('lb', 'ton')
    _min = 0
    _cumulative = True
    _abbreviation = 'm'

    def _kg_to_lb(self, value):
        return value * 2.20462

    def _kg_to_g(self, value):
        return value * 1000.

    def _kg_to_tonne(self, value):
        return value / 1000.

    def _kg_to_ton(self, value):
        return value / 907.185

    def _kg_to_oz(self, value):
        return value * 35.274

    def _lb_to_kg(self, value):
        return value / 2.20462

    def _g_to_kg(self, value):
        return value / 1000.

    def _tonne_to_kg(self, value):
        return value * 1000.

    def _ton_to_kg(self, value):
        return value * 907.185

    def _oz_to_kg(self, value):
        return value / 35.274

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('kg', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'tonne':
            return self.to_unit(values, 'ton', from_unit), 'ton'
        else:
            return self.to_unit(values, 'lb', from_unit), 'lb'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'ton':
            return self.to_unit(values, 'tonne', from_unit), 'tonne'
        else:
            return self.to_unit(values, 'kg', from_unit), 'kg'



# coding=utf-8
"""Volumetric Heat Capacity data type."""
from __future__ import division

from .base import DataTypeBase


class VolumetricHeatCapacity(DataTypeBase):
    """Volumetric Heat Capacity"""
    _units = ('J/m3-K', 'Btu/ft3-F', 'kWh/m3-K', 'kBtu/ft3-F', 'kJ/m3-K', 'MJ/m3-K')
    _si_units = ('J/m3-K', 'kWh/m3-K', 'kJ/m3-K', 'MJ/m3-K')
    _ip_units = ('Btu/ft3-F', 'kBtu/ft3-F')
    _min = 0
    _abbreviation = 'c'

    def _J_m3K_to_Btu_ft3F(self, value):
        return value / 67066.1

    def _J_m3K_to_kWh_m3K(self, value):
        return value / 3600000.

    def _J_m3K_to_kBtu_ft3F(self, value):
        return value / 67066100.

    def _J_m3K_to_kJ_m3K(self, value):
        return value / 1000.

    def _J_m3K_to_MJ_m3K(self, value):
        return value / 1000000.

    def _Btu_ft3F_to_J_m3K(self, value):
        return value * 67066.1

    def _kWh_m3K_to_J_m3K(self, value):
        return value * 3600000.

    def _kBtu_ft3F_to_J_m3K(self, value):
        return value * 67066100.

    def _kJ_m3K_to_J_m3K(self, value):
        return value * 1000.

    def _MJ_m3K_to_J_m3K(self, value):
        return value * 1000000.

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('J/m3-K', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'kWh/m3-K':
            return self.to_unit(values, 'kBtu/ft3-F', from_unit), 'kBtu/ft3-F'
        else:
            return self.to_unit(values, 'Btu/ft3-F', from_unit), 'Btu/ft3-F'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'kBtu/ft3-F':
            return self.to_unit(values, 'kWh/m3-K', from_unit), 'kWh/m3-K'
        else:
            return self.to_unit(values, 'J/m3-K', from_unit), 'J/m3-K'



# coding=utf-8
"""Utilities for finding roots of continuous functions."""
from __future__ import division


def secant(a, b, fn, epsilon):
    """Solve for a root using one of the fastest root-finding algorithms.

    The method calculates the slope of the function fn and this enables it to converge
    to a solution very fast. However, if started too far away from a root, the method
    may not converge (returning a None). For this reason, it is recommended that this
    function be used first in any guess-and-check workflow and, if it fails to find a
    root, the bisect() method should be used.

    Args:
        a: The lowest possible boundary of the value you are tying to find.
        b: The highest possible boundary of the value you are tying to find.
        fn: A function representing the relationship between the value you are
            trying to find and the target condition you are trying to satisfy.
            It should typically be structured in the following way:

            .. code-block:: python

                def fn(value_trying_to_find):
                    funct(value_trying_to_find) - target_desired_from_funct

            ...but the subtraction should be switched if value_trying_to_find
            has a negative relationship with the funct.
        epsilon: The acceptable error in the target_desired_from_funct.

    Returns:
        root -- The value that gives the target_desired_from_funct in the sample
        above (aka. the value that returns 0 from the fn).

    References
    ----------
    [1] Wikipedia contributors. (2018, December 29). Root-finding algorithm.
    In Wikipedia, The Free Encyclopedia. Retrieved 18:16, December 30, 2018,
    from https://en.wikipedia.org/wiki/Root-finding_algorithm#Secant_method
    """
    f1 = fn(a)
    if abs(f1) <= epsilon:
        return a
    f2 = fn(b)
    if abs(f2) <= epsilon:
        return b
    for _ in range(100):
        slope = (f2 - f1) / (b - a)
        c = b - f2 / slope
        f3 = fn(c)
        if abs(f3) < epsilon:
            return c
        a = b
        b = c
        f1 = f2
        f2 = f3
    return None


def bisect(a, b, fn, epsilon, target=0):
    """Solve for a root using the simplest root-finding algorithm.

    It is extremely reliable. However, it converges slowly for this reason,
    it is recommended that this only be used after the secant() method has
    returned None.

    Args:
        a: A lower guess of the value you are tying to find.
        b: A higher guess of the value you are tying to find.
        fn: A function representing the relationship between the value you are
            trying to find and the target condition you are trying to satisfy.
            It should typically be structured in the following way:

            .. code-block:: python

                def fn(value_trying_to_find):
                    funct(value_trying_to_find) - target_desired_from_funct

            ...but the subtraction should be switched if value_trying_to_find
            has a negative relationship with the funct.
        epsilon: The acceptable error in the target_desired_from_funct.
        target: The target slope (typically 0 for a local minima or maxima).

    Returns:
        root -- The value that gives the target_desired_from_funct.

    References
    ----------
    [1] Wikipedia contributors. (2018, December 29). Root-finding algorithm.
    In Wikipedia, The Free Encyclopedia. Retrieved 18:16, December 30, 2018,
    from https://en.wikipedia.org/wiki/Root-finding_algorithm#Bisection_method
    """
    max_e = 2 * epsilon
    while (abs(b - a) > max_e):
        midpoint = (b + a) / 2
        a_t = fn(a)
        b_t = fn(b)
        midpoint_t = fn(midpoint)
        if (a_t - target) * (midpoint_t - target) < 0:
            b = midpoint
        elif (b_t - target) * (midpoint_t - target) < 0:
            a = midpoint
        else:
            return midpoint
    return midpoint


def secant_three_var(a, b, fn, epsilon, other_args):
    """Solve the roots of a 3-variable function with one of the fastest algorithms.

    Args:
        a: A tuple with 3 numbers for the the lowest boundary of the roots.
        b: A tuple with 3 numbers for the the highest boundary of the roots.
        fn: A function for which roots are to be solved. That is, where the output
            of the function is a tuple of three zeros.
        epsilon: The acceptable error in the resulting root.
        other_args: Other input arguments for the fn other than the ones being
            adjusted to solve the root.

    Returns:
        root -- a tuple of 3 values that return a vector of zeros from the fn.
    """
    args_1 = (a,) + other_args
    f1 = fn(*args_1)
    if all(abs(v) <= epsilon for v in f1):
        return a
    args_2 = (b,) + other_args
    f2 = fn(*args_2)
    if all(abs(v) <= epsilon for v in f2):
        return b

    for _ in range(1000):
        try:
            slope = tuple((v2 - v1) / (bv - av) for v1, v2, av, bv in zip(f1, f2, a, b))
        except ZeroDivisionError:
            return None  # failed to converge
        try:
            c = tuple(bv - v2 / s for bv, v2, s in zip(b, f2, slope))
        except ZeroDivisionError:  # zero slope was found
            return None  # failed to converge
        args_3 = (c,) + other_args
        f3 = fn(*args_3)
        if all(abs(v) <= epsilon for v in f3):
            return c
        a = b
        b = c
        f1 = f2
        f2 = f3

    return None



# coding=utf-8
from __future__ import division

from .location import Location

from .dt import DateTime, Date
from .header import Header
from .analysisperiod import AnalysisPeriod
from .datacollection import HourlyContinuousCollection
from .sunpath import Sunpath

from .datatype import angle, energyflux, energyintensity, \
    fraction, pressure, speed, temperature

from .skymodel import ashrae_revised_clear_sky, \
    ashrae_clear_sky, calc_horizontal_infrared

from .psychrometrics import rel_humid_from_db_dpt, dew_point_from_db_hr, \
    dew_point_from_db_enth, dew_point_from_db_wb

import math
import re
import sys
if (sys.version_info > (3, 0)):
    xrange = range


class DesignDay(object):
    """An object representing design day conditions.

    Args:
        name: A text string to set the name of the design day
        day_type: Choose from 'SummerDesignDay', 'WinterDesignDay' or other
            EnergyPlus days visible under the DAY_TYPES property of this object.
        location: Location object for the design day.
        dry_bulb_condition: Ladybug DryBulbCondition object.
        humidity_condition: Ladybug HumidityCondition object.
        wind_condition: Ladybug WindCondition object.
        sky_condition: Ladybug SkyCondition object (either ASHRAEClearSky or ASHRAETau).

    Properties:
        * name
        * day_type
        * location
        * dry_bulb_condition
        * humidity_condition
        * wind_condition
        * sky_condition
        * analysis_period
        * hourly_datetimes
        * hourly_dry_bulb
        * hourly_dew_point
        * hourly_relative_humidity
        * hourly_barometric_pressure
        * hourly_wind_speed
        * hourly_wind_direction
        * hourly_solar_radiation
        * hourly_sky_cover
        * hourly_horizontal_infrared
    """
    __slots__ = ('_name', '_day_type', '_location', '_dry_bulb_condition',
                 '_humidity_condition', '_wind_condition', '_sky_condition')

    # possible day types
    DAY_TYPES = ('SummerDesignDay', 'WinterDesignDay', 'Sunday', 'Monday',
                 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Holiday',
                 'CustomDay1', 'CustomDay2')

    # keys denoting the values from which design days are derived
    # these keys and their order come from Climate Design Data of ASHRAE Handbook
    # https://ashrae-meteo.info/
    HEATING_KEYS = ('Month', 'DB996', 'DB990', 'DP996', 'HR_DP996', 'DB_DP996',
                    'DP990', 'HR_DP990', 'DB_DP990', 'WS004c', 'DB_WS004c',
                    'WS010c', 'DB_WS010c', 'WS_DB996', 'WD_DB996', 'WSF')
    COOLING_KEYS = ('Month', 'DBR', 'DB004', 'WB_DB004', 'DB010', 'WB_DB010',
                    'DB020', 'WB_DB020', 'WB004', 'DB_WB004', 'WB010', 'DB_WB010',
                    'WB020', 'DB_WB020', 'WS_DB004', 'WD_DB004', 'DP004',
                    'HR_DP004', 'DB_DP004', 'DP010', 'HR_DP010', 'DB_DP010',
                    'DP020', 'HR_DP020', 'DB_DP020', 'EN004', 'DB_EN004',
                    'EN010', 'DB_EN010', 'EN020', 'DB_EN020', 'Hrs_8-4_&_DB')
    EXTREME_KEYS = ('WS010', 'WS025', 'WS050', 'WBmax', 'DBmin_mean', 'DBmax_mean',
                    'DBmin_stddev', 'DBmax_stddev', 'DBmin05years', 'DBmax05years',
                    'DBmin10years', 'DBmax10years', 'DBmin20years', 'DBmax20years',
                    'DBmin50years', 'DBmax50years')

    # comments that go along with the IDF string representation of the design day
    IDF_COMMENTS = ('!- Name',
                    '!- Month',
                    '!- Day of Month',
                    '!- Day Type',
                    '!- Max Dry-Bulb Temp {C}',
                    '!- Daily Dry-Bulb Temp Range {C}',
                    '!- Dry-Bulb Temp Range Modifier Type',
                    '!- Dry-Bulb Temp Range Modifier Schedule Name',
                    '!- Humidity Condition Type',
                    '!- Wetbulb/Dewpoint at Max Dry-Bulb {C}',
                    '!- Humidity Indicating Day Schedule Name',
                    '!- Humidity Ratio at Maximum Dry-Bulb {kgWater/kgDryAir}',
                    '!- Enthalpy at Maximum Dry-Bulb {J/kg}',
                    '!- Daily Wet-Bulb Temperature Range {deltaC}',
                    '!- Barometric Pressure {Pa}',
                    '!- Wind Speed {m/s}',
                    '!- Wind Direction {Degrees; N=0, S=180}',
                    '!- Rain {Yes/No}',
                    '!- Snow on ground {Yes/No}',
                    '!- Daylight Savings Time Indicator {Yes/No}',
                    '!- Solar Model Indicator',
                    '!- Beam Solar Day Schedule Name',
                    '!- Diffuse Solar Day Schedule Name',
                    '!- ASHRAE Clear Sky Beam Optical Depth (taub)',
                    '!- ASHRAE Clear Sky Diffuse Optical Depth (taud)',
                    '!- Clearness (0.0 to 1.2)')

    def __init__(self, name, day_type, location, dry_bulb_condition, humidity_condition,
                 wind_condition, sky_condition):
        self.name = str(name)
        self.day_type = day_type
        self.location = location
        self.dry_bulb_condition = dry_bulb_condition
        self.humidity_condition = humidity_condition
        self.wind_condition = wind_condition
        self.sky_condition = sky_condition

    @classmethod
    def from_dict(cls, data):
        """Create a Design Day from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "DesignDay",
            "name": "",  # string
            "day_type": "",  # string
            "location": {},  # optional ladybug Location dictionary
            "dry_bulb_condition": {},  # ladybug DryBulbCondition dictionary
            "humidity_condition": {},  # ladybug HumidityCondition dictionary
            "wind_condition": {},  # ladybug WindCondition dictionary
            "sky_condition": {}  # ladybug ASHRAEClearSky or ASHRAETau dictionary
            }
        """
        required_keys = ('name', 'day_type', 'dry_bulb_condition',
                         'humidity_condition', 'wind_condition', 'sky_condition')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        if 'location' in data and data['location'] is not None:
            loc = Location.from_dict(data['location'])
        else:
            loc = Location()

        return cls(data['name'], data['day_type'], loc,
                   DryBulbCondition.from_dict(data['dry_bulb_condition']),
                   HumidityCondition.from_dict(data['humidity_condition']),
                   WindCondition.from_dict(data['wind_condition']),
                   _SkyCondition.from_dict(data['sky_condition']))

    @classmethod
    def from_idf(cls, idf_string, location):
        """Initialize from an EnergyPlus IDF string of a SizingPeriod:DesignDay.

        Args:
            idf_string: A full IDF string representing a SizingPeriod:DesignDay.
            location: A Ladybug Location object, used to interpret the sky condition
                over the course of the design day.
        """
        # format the object into a list of properties
        idf_string = idf_string.strip()
        assert idf_string.startswith('SizingPeriod:DesignDay'), 'Expected SizingPeriod' \
            ':DesignDay but received a different object: {}'.format(idf_string)
        idf_string = idf_string.replace(';', ',')
        idf_string = re.sub(r'!.*\n', '', idf_string)
        ep_fields = [e_str.strip() for e_str in idf_string.split(',')]

        # extract primary properties
        name = ep_fields[1]
        day_type = ep_fields[4]

        # extract dry bulb temperatures
        dry_bulb_condition = DryBulbCondition(
            float(ep_fields[5]), float(ep_fields[6]), ep_fields[7], ep_fields[8])

        # extract humidity conditions
        h_type = ep_fields[9]
        h_val = 0 if ep_fields[10] == '' else float(ep_fields[10])
        rain = True if len(ep_fields) > 18 and ep_fields[18].lower() == 'yes' else False
        snow = True if len(ep_fields) > 19 and ep_fields[19].lower() == 'yes' else False
        if h_type == 'HumidityRatio':
            h_val = float(ep_fields[12])
        elif h_type == 'Enthalpy':
            h_val = float(ep_fields[13])
        humidity_condition = HumidityCondition(
            h_type, h_val, float(ep_fields[15]), rain, snow, ep_fields[11])

        # extract wind conditions
        wind_condition = WindCondition(float(ep_fields[16]), float(ep_fields[17]))

        # extract the sky conditions
        date_obj = Date(int(ep_fields[2]), int(ep_fields[3]))
        dl_save = True if len(ep_fields) > 20 and ep_fields[20].lower() == 'yes' \
            else False
        if len(ep_fields) > 21:
            sky_model = ep_fields[21]
            if sky_model == 'ASHRAEClearSky':
                sky_clr = float(ep_fields[26]) if len(ep_fields) > 26 else 0
                sky_condition = ASHRAEClearSky(date_obj, sky_clr, dl_save)
            elif sky_model in ('ASHRAETau', 'ASHRAETau2017'):
                use_2017 = True if sky_model.endswith('2017') else False
                t_b = float(ep_fields[24]) if len(ep_fields) > 24 else 0
                t_d = float(ep_fields[25]) if len(ep_fields) > 25 else 0
                sky_condition = ASHRAETau(date_obj, t_b, t_d, use_2017, dl_save)
            else:
                sky_condition = _SkyCondition(date_obj, dl_save)
            if sky_model == 'Schedule':
                sky_condition.beam_schedule = ep_fields[22]
                sky_condition.diffuse_schedule = ep_fields[23]
        else:  # default of no sky condition
            sky_condition = ASHRAEClearSky(date_obj, 0, dl_save)

        return cls(name, day_type, location, dry_bulb_condition,
                   humidity_condition, wind_condition, sky_condition)

    @classmethod
    def from_design_day_properties(cls, name, day_type, location, date,
                                   dry_bulb_max, dry_bulb_range, humidity_type,
                                   humidity_value, pressure, wind_speed, wind_dir,
                                   sky_model, sky_properties):
        """Create a design day object from various key properties.

        Args:
            name: A text string to set the name of the design day
            day_type: Choose from 'SummerDesignDay', 'WinterDesignDay' or other
                EnergyPlus days
            location: Location for the design day
            date: Ladybug Date object for the design day.
            dry_bulb_max: Maximum dry bulb temperature over the design day (in C).
            dry_bulb_range: Dry bulb range over the design day (in C).
            humidity_type: Type of humidity to use. Choose from
                Wetbulb, Dewpoint, HumidityRatio, Enthalpy
            humidity_value: The value of the condition above.
            pressure: Barometric pressure in Pa.
            wind_speed: Wind speed over the design day in m/s.
            wind_dir: Wind direction over the design day in degrees.
            sky_model: Type of solar model to use. Choose from ASHRAEClearSky, ASHRAETau.
            sky_properties: A list of properties describing the sky above.
                For ASHRAEClearSky this is a single value for clearness
                For ASHRAETau, this is the tau_beam and tau_diffuse
        """
        dry_bulb_condition = DryBulbCondition(dry_bulb_max, dry_bulb_range)
        humidity_condition = HumidityCondition(
            humidity_type, humidity_value, pressure)
        wind_condition = WindCondition(wind_speed, wind_dir)
        if sky_model == 'ASHRAEClearSky':
            sky_condition = ASHRAEClearSky(date, sky_properties[0])
        elif sky_model in ('ASHRAETau', 'ASHRAETau2017'):
            use_2017 = True if sky_model.endswith('2017') else False
            sky_condition = ASHRAETau(
                date, sky_properties[0], sky_properties[-1], use_2017)
        return cls(name, day_type, location, dry_bulb_condition,
                   humidity_condition, wind_condition, sky_condition)

    @classmethod
    def from_ashrae_dict_heating(cls, ashrae_dict, location,
                                 use_990=False, pressure=None):
        """Create a heating design day object from a ASHRAE HOF dictionary.

        Args:
            ashrae_dict: A dictionary with 15 keys that match those in the
                HEATING_KEYS property of this object. Each key should
                correspond to a value.
            location: Location object for the design day
            use_990: Boolean to denote what type of design day to create
                (wether it is a 99.0% or a 99.6% design day).  Default is
                False for a 99.6% annual heating design day
            pressure: Atmospheric pressure in Pa that should be used in the
                creation of the humidity condition. Default is 101325 Pa
                for pressure at sea level.
        """
        db_key = 'DB996' if not use_990 else 'DB990'
        perc_str = '99.6' if not use_990 else '99'
        pressure = pressure if pressure is not None else 101325
        db_cond = DryBulbCondition(float(ashrae_dict[db_key]), 0)
        hu_cond = HumidityCondition('Wetbulb', float(ashrae_dict[db_key]), pressure)
        ws_cond = WindCondition(float(ashrae_dict['WS_DB996']),
                                float(ashrae_dict['WD_DB996']))
        date_obj = Date(int(ashrae_dict['Month']), 21)
        sky_cond = ASHRAEClearSky(date_obj, 0)
        name = '{} Heating Design Day {}% Condns DB'.format(location.city, perc_str)
        return cls(name, 'WinterDesignDay', location,
                   db_cond, hu_cond, ws_cond, sky_cond)

    @classmethod
    def from_ashrae_dict_cooling(cls, ashrae_dict, location,
                                 use_010=False, pressure=None, tau=None):
        """Create a heating design day object from a ASHRAE HOF dictionary.

        Args:
            ashrae_dict: A dictionary with 32 keys that match those in the
                COOLING_KEYS property of this object. Each key should
                correspond to a value.
            location: Location object for the design day
            use_010: Boolean to denote what type of design day to create
                (wether it is a 1.0% or a 0.4% design day).  Default is
                False for a 0.4% annual cooling design day
            pressure: Atmospheric pressure in Pa that should be used in the
                creation of the humidity condition. Default is 101325 Pa
                for pressure at sea level.
            tau: Optional tuple containing two values, which will set the
                sky condition to be a revised ASHRAE clear sky (Tau model).
                The first item of the tuple should be the tau beam value
                and the second is for the tau diffuse value.  Default is
                None, which will default to the original ASHRAE Clear Sky.
        """
        db_key = 'DB004' if not use_010 else 'DB010'
        wb_key = 'WB_DB004' if not use_010 else 'WB_DB010'
        perc_str = '0.4' if not use_010 else '1'
        pressure = pressure if pressure is not None else 101325
        db_cond = DryBulbCondition(float(ashrae_dict[db_key]), float(ashrae_dict['DBR']))
        hu_cond = HumidityCondition('Wetbulb', float(ashrae_dict[wb_key]), pressure)
        ws_cond = WindCondition(float(ashrae_dict['WS_DB004']),
                                float(ashrae_dict['WD_DB004']))
        date_obj = Date(int(ashrae_dict['Month']), 21)
        if tau is not None:
            sky_cond = ASHRAETau(date_obj, tau[0], tau[1])
        else:
            sky_cond = ASHRAEClearSky(date_obj)
        name = '{} Cooling Design Day {}% Condns DB=>MWB'.format(
            location.city, perc_str)
        return cls(name, 'SummerDesignDay', location,
                   db_cond, hu_cond, ws_cond, sky_cond)

    @property
    def name(self):
        """Get or set the name."""
        return self._name

    @name.setter
    def name(self, data):
        assert isinstance(data, str), 'Design day name must be' \
            ' text. Got {}'.format(type(data))
        self._name = data

    @property
    def day_type(self):
        """Get or set the type of design day."""
        return self._day_type

    @day_type.setter
    def day_type(self, data):
        assert data in self.DAY_TYPES, 'day_type {} is not' \
            ' recognized'.format(data)
        self._day_type = data

    @property
    def location(self):
        """Get or set the location."""
        return self._location

    @location.setter
    def location(self, data):
        assert isinstance(data, Location), 'Expected' \
            ' Location type. Got {}'.format(type(data))
        self._location = data

    @property
    def dry_bulb_condition(self):
        """Get or set the dry bulb conditions."""
        return self._dry_bulb_condition

    @dry_bulb_condition.setter
    def dry_bulb_condition(self, data):
        assert isinstance(data, DryBulbCondition), 'Expected' \
            ' DryBulbCondition type. Got {}'.format(type(data))
        self._dry_bulb_condition = data

    @property
    def humidity_condition(self):
        """Get or set the humidity conditions."""
        return self._humidity_condition

    @humidity_condition.setter
    def humidity_condition(self, data):
        assert isinstance(data, HumidityCondition), 'Expected' \
            ' HumidityCondition type. Got {}'.format(type(data))
        self._humidity_condition = data

    @property
    def wind_condition(self):
        """Get or set the wind conditions."""
        return self._wind_condition

    @wind_condition.setter
    def wind_condition(self, data):
        assert isinstance(data, WindCondition), 'Expected' \
            ' WindCondition type. Got {}'.format(type(data))
        self._wind_condition = data

    @property
    def sky_condition(self):
        """Get or set the sky conditions."""
        return self._sky_condition

    @sky_condition.setter
    def sky_condition(self, data):
        assert isinstance(data, _SkyCondition), 'Expected ASHRAEClearSky or' \
            ' ASHRAETau type. Got {}'.format(type(data))
        self._sky_condition = data

    @property
    def analysis_period(self):
        """Get the analysisperiod of the design day."""
        return AnalysisPeriod(
            self.sky_condition.date.month,
            self.sky_condition.date.day,
            0,
            self.sky_condition.date.month,
            self.sky_condition.date.day,
            23)

    @property
    def hourly_datetimes(self):
        """Get a list of hourly DateTime objects for the DesignDay."""
        start_moy = self.sky_condition.date.doy * 1440
        lp_yr = self.sky_condition.date.leap_year
        return tuple(DateTime.from_moy(start_moy + (i * 60), lp_yr) for i in xrange(24))

    @property
    def hourly_dry_bulb(self):
        """A data collection containing hourly dry bulb temperature over they day."""
        return self._get_daily_data_collections(
            temperature.DryBulbTemperature(), 'C',
            self._dry_bulb_condition.hourly_values)

    @property
    def hourly_dew_point(self):
        """A data collection containing hourly dew points over they day."""
        dpt_data = self._humidity_condition.hourly_dew_point_values(
            self._dry_bulb_condition)
        return self._get_daily_data_collections(
            temperature.DewPointTemperature(), 'C', dpt_data)

    @property
    def hourly_relative_humidity(self):
        """A data collection containing hourly relative humidity over they day."""
        dpt_data = self._humidity_condition.hourly_dew_point_values(
            self._dry_bulb_condition)
        rh_data = [rel_humid_from_db_dpt(x, y) for x, y in zip(
            self._dry_bulb_condition.hourly_values, dpt_data)]
        return self._get_daily_data_collections(
            fraction.RelativeHumidity(), '%', rh_data)

    @property
    def hourly_barometric_pressure(self):
        """A data collection containing hourly barometric pressures over they day."""
        return self._get_daily_data_collections(
            pressure.AtmosphericStationPressure(), 'Pa',
            self._humidity_condition.hourly_pressure)

    @property
    def hourly_wind_speed(self):
        """A data collection containing hourly wind speeds over they day."""
        return self._get_daily_data_collections(
            speed.WindSpeed(), 'm/s', self._wind_condition.hourly_values)

    @property
    def hourly_wind_direction(self):
        """A data collection containing hourly wind directions over they day."""
        return self._get_daily_data_collections(
            angle.WindDirection(), 'degrees', self._wind_condition.hourly_wind_dirs)

    @property
    def hourly_solar_radiation(self):
        """Three data collections containing hourly direct normal, diffuse horizontal,
        and global horizontal radiation.
        """
        dir_norm, diff_horiz, glob_horiz = \
            self._sky_condition.radiation_values(self._location)

        dir_norm_data = self._get_daily_data_collections(
            energyintensity.DirectNormalRadiation(), 'Wh/m2', dir_norm)
        diff_horiz_data = self._get_daily_data_collections(
            energyintensity.DiffuseHorizontalRadiation(), 'Wh/m2', diff_horiz)
        glob_horiz_data = self._get_daily_data_collections(
            energyintensity.GlobalHorizontalRadiation(), 'Wh/m2', glob_horiz)

        return dir_norm_data, diff_horiz_data, glob_horiz_data

    @property
    def hourly_sky_cover(self):
        """A data collection containing hourly sky cover values in tenths."""
        return self._get_daily_data_collections(
            fraction.TotalSkyCover(), 'tenths', self._sky_condition.hourly_sky_cover)

    @property
    def hourly_horizontal_infrared(self):
        """A data collection containing hourly horizontal infrared intensity in W/m2.
        """
        sky_cover = self._sky_condition.hourly_sky_cover
        db_temp = self._dry_bulb_condition.hourly_values
        dp_temp = self._humidity_condition.hourly_dew_point_values(
            self._dry_bulb_condition)

        horiz_ir = []
        for i in xrange(len(sky_cover)):
            horiz_ir.append(
                calc_horizontal_infrared(sky_cover[i], db_temp[i], dp_temp[i]))

        return self._get_daily_data_collections(
            energyflux.HorizontalInfraredRadiationIntensity(), 'W/m2', horiz_ir)

    def to_idf(self):
        """Get this object as an EnergyPlus IDF SizingPeriod:DesignDay string."""
        # Put together the values in the order that they exist in the ddy file
        ep_vals = [self.name,
                   self.sky_condition.date.month,
                   self.sky_condition.date.day,
                   self.day_type,
                   self.dry_bulb_condition.dry_bulb_max,
                   self.dry_bulb_condition.dry_bulb_range,
                   self.dry_bulb_condition.modifier_type,
                   self.dry_bulb_condition.modifier_schedule,
                   self.humidity_condition.humidity_type, '',
                   self.humidity_condition.schedule, '', '',
                   self.humidity_condition.wet_bulb_range,
                   self.humidity_condition.barometric_pressure,
                   self.wind_condition.wind_speed,
                   self.wind_condition.wind_direction,
                   'Yes' if self.humidity_condition.rain else 'No',
                   'Yes' if self.humidity_condition.snow_on_ground else 'No',
                   'Yes' if self.sky_condition.daylight_savings else 'No',
                   'ASHRAEClearSky',
                   self.sky_condition.beam_schedule,
                   self.sky_condition.diffuse_schedule, '', '', '']

        # assign humidity values based on the type of criteria
        if self.humidity_condition.humidity_type == 'Wetbulb' or \
                self.humidity_condition.humidity_type == 'Dewpoint':
            ep_vals[9] = self.humidity_condition.humidity_value
        elif self.humidity_condition.humidity_type == 'HumidityRatio':
            ep_vals[11] = self.humidity_condition.humidity_value
        elif self.humidity_condition.humidity_type == 'Enthalpy':
            ep_vals[12] = self.humidity_condition.humidity_value

        # assign sky condition values based on the solar model
        if isinstance(self.sky_condition, ASHRAEClearSky):
            ep_vals[25] = self.sky_condition.clearness
        if isinstance(self.sky_condition, ASHRAETau):
            ep_vals[20] = 'ASHRAETau2017' if self.sky_condition.use_2017 else 'ASHRAETau'
            ep_vals[23] = self.sky_condition.tau_b
            ep_vals[24] = self.sky_condition._tau_d
            ep_vals.pop()

        # put everything together into one string
        commented_str = ['  {},{}{}\n'.format(
            str(val), ' ' * (60 - len(str(val))), self.IDF_COMMENTS[i])
            for i, val in enumerate(ep_vals)]
        commented_str[-1] = commented_str[-1].replace(',', ';')
        commented_str.insert(0, 'SizingPeriod:DesignDay,\n')
        commented_str.append('\n')

        return ''.join(commented_str)

    def to_dict(self, include_location=True):
        """Convert the Design Day to a dictionary.

        Args:
            include_location: If True, the location dictionary will be included in the
                returned dictionary. This ensures that the sky condition contained
                in any re-serialized design day can produce accurate radiation values.
                However, this location is not necessary when exporting design days
                for energy simulation since the EPW simulated with the model already
                contains location information. Default: True.
        """
        des_day_dict = {
            'type': 'DesignDay',
            'name': self.name,
            'day_type': self.day_type,
            'dry_bulb_condition': self.dry_bulb_condition.to_dict(),
            'humidity_condition': self.humidity_condition.to_dict(),
            'wind_condition': self.wind_condition.to_dict(),
            'sky_condition': self.sky_condition.to_dict(),
            'type': 'DesignDay'
        }
        if include_location:
            des_day_dict['location'] = self.location.to_dict()
        return des_day_dict

    def _get_daily_data_collections(self, data_type, unit, values):
        """Return an empty data collection."""
        data_header = Header(data_type=data_type, unit=unit,
                             analysis_period=self.analysis_period,
                             metadata={'source': self._location.source,
                                       'country': self._location.country,
                                       'city': self._location.city})
        return HourlyContinuousCollection(data_header, values)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __copy__(self):
        return DesignDay(
            self._name, self._day_type, self._location.duplicate(),
            self._dry_bulb_condition.duplicate(), self._humidity_condition.duplicate(),
            self._wind_condition.duplicate(), self._sky_condition.duplicate())

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._name, self._day_type, hash(self._location),
                hash(self._dry_bulb_condition), hash(self._humidity_condition),
                hash(self._wind_condition), hash(self._sky_condition))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, DesignDay) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Design day representation."""
        return "Design Day - {} [{}]".format(self.name, self._day_type)


class DryBulbCondition(object):
    """Represents dry bulb conditions on a design day.

    Args:
        dry_bulb_max: The maximum dry bulb temperature on the design day [C].
        dry_bulb_range: The difference between min and max temperatures on the
            design day [C].
        modifier_type: Optional string for the type of modifier used to estimate
            temperature at a given timestep. Choose from the following:
            'DefaultMultipliers', 'MultiplierSchedule', 'DifferenceSchedule',
            'TemperatureProfileSchedule'. Default: 'DefaultMultipliers'
        modifier_schedule: Optional text string for the name of the modifier
            schedule. Should be an empty string unless 'MultiplierSchedule' is
            selected as the modifier_type.

    Properties:
        * dry_bulb_max
        * dry_bulb_range
        * modifier_type
        * modifier_schedule
        * hourly_values
    """
    __slots__ = ('_dry_bulb_max', '_dry_bulb_range', 'modifier_type',
                 'modifier_schedule')
    HOURLY_MULTIPLIERS = (0.82, 0.88, 0.92, 0.95, 0.98, 1, 0.98, 0.91, 0.74, 0.55,
                          0.38, 0.23, 0.13, 0.05, 0, 0, 0.06, 0.14, 0.24, 0.39, 0.5,
                          0.59, 0.68, 0.75)

    def __init__(self, dry_bulb_max, dry_bulb_range,
                 modifier_type='DefaultMultipliers', modifier_schedule=''):
        self.dry_bulb_max = dry_bulb_max
        self.dry_bulb_range = dry_bulb_range
        self.modifier_type = str(modifier_type)
        self.modifier_schedule = str(modifier_schedule)

    @classmethod
    def from_dict(cls, data):
        """Create a Dry Bulb Condition from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "DryBulbCondition",
            "dry_bulb_max": 0.0,  # float
            "dry_bulb_range": 0.0  # float
            }
        """
        # check required keys
        required_keys = ('dry_bulb_max', 'dry_bulb_range')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        # set defaults for optional keys
        mod_type = data['modifier_type'] if 'modifier_type' in data \
            else 'DefaultMultipliers'
        mod_schedule = data['modifier_schedule'] if 'modifier_schedule' in data else ''

        return cls(data['dry_bulb_max'], data['dry_bulb_range'], mod_type, mod_schedule)

    @property
    def hourly_values(self):
        """Get a list of temperature values for each hour over the design day."""
        return [self._dry_bulb_max - self._dry_bulb_range * x for
                x in self.HOURLY_MULTIPLIERS]

    @property
    def dry_bulb_max(self):
        """Get or set the max dry bulb temperature."""
        return self._dry_bulb_max

    @dry_bulb_max.setter
    def dry_bulb_max(self, data):
        assert isinstance(data, (float, int)), 'dry_bulb_max must be a' \
            ' number. Got {}'.format(type(data))
        self._dry_bulb_max = data

    @property
    def dry_bulb_range(self):
        """Get or set the dry bulb range over the day."""
        return self._dry_bulb_range

    @dry_bulb_range.setter
    def dry_bulb_range(self, data):
        assert isinstance(data, (float, int)), 'dry_bulb_range must be a' \
            ' number. Got {}'.format(type(data))
        assert data >= 0, 'dry_bulb_range must be greater than or equal to' \
            ' zero. Got {}'.format(data)
        self._dry_bulb_range = data

    def to_dict(self):
        """Convert the Dry Bulb Condition to a dictionary."""
        return {
            'type': 'DryBulbCondition',
            'dry_bulb_max': self.dry_bulb_max,
            'dry_bulb_range': self.dry_bulb_range
        }

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __copy__(self):
        return DryBulbCondition(self._dry_bulb_max, self._dry_bulb_range,
                                self.modifier_type, self.modifier_schedule)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._dry_bulb_max, self._dry_bulb_range,
                self.modifier_type, self.modifier_schedule)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, DryBulbCondition) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Dry bulb condition representation."""
        return "Dry Bulb Condition [Max: {}, Range: {}]".format(
            str(self._dry_bulb_max), str(self._dry_bulb_range))


class HumidityCondition(object):
    """Represents humidity and precipitation conditions on the design day.

    Args:
        humidity_type: Choose from: Wetbulb, Dewpoint, HumidityRatio, Enthalpy
        humidity_value: The value correcponding to the humidity_type
        barometric_pressure: Default is to use pressure at sea level
        rain: Boolean to indicate rain on the design day. Default: False.
        snow_on_ground: Boolean to indicate snow on the design day. Default: False.
        schedule: Optional humidity schedule
        wet_bulb_range: Optional wet bulb temperature range

    Properties:
        * humidity_type
        * humidity_value
        * barometric_pressure
        * rain
        * snow_on_ground
        * schedule
        * wet_bulb_range
        * hourly_pressure
    """
    __slots__ = ('_humidity_type', '_humidity_value', '_barometric_pressure',
                 '_rain', '_snow_on_ground', 'schedule', 'wet_bulb_range')
    HUMIDITY_TYPES = ('Wetbulb', 'Dewpoint', 'HumidityRatio', 'Enthalpy')

    def __init__(self, humidity_type, humidity_value, barometric_pressure=101325,
                 rain=False, snow_on_ground=False, schedule='', wet_bulb_range=''):
        self.humidity_type = humidity_type
        self.humidity_value = humidity_value
        self.barometric_pressure = barometric_pressure
        self.rain = rain
        self.snow_on_ground = snow_on_ground
        self.schedule = schedule
        self.wet_bulb_range = wet_bulb_range

    @classmethod
    def from_dict(cls, data):
        """Create a Humidity Condition from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "HumidityCondition",
            "humidity_type": "",  # string
            "humidity_value": 0.0,  # float
            "barometric_pressure": 0.0,  # float
            "rain": False,  # bool
            "snow_on_ground": False  # bool
            }
        """
        # check required keys
        required_keys = ('humidity_type', 'humidity_value')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        # set defaults for optional keys
        press = data['barometric_pressure'] if 'barometric_pressure' in data else 101325
        rain = data['rain'] if 'rain' in data else False
        snow_on_ground = data['snow_on_ground'] if 'snow_on_ground' in data else False
        sched = data['schedule'] if 'schedule' in data else ''
        wet_bulb_range = data['wet_bulb_range'] if 'wet_bulb_range' in data else ''

        return cls(data['humidity_type'], data['humidity_value'], press,
                   rain, snow_on_ground, sched, wet_bulb_range)

    @property
    def humidity_type(self):
        """Get or set the humidity condition type."""
        return self._humidity_type

    @humidity_type.setter
    def humidity_type(self, data):
        assert data in self.HUMIDITY_TYPES, 'humidity_type {} is not' \
            ' recognized'.format(data)
        self._humidity_type = data

    @property
    def humidity_value(self):
        """Get or set the value of the humidity."""
        return self._humidity_value

    @humidity_value.setter
    def humidity_value(self, data):
        assert isinstance(data, (float, int)), 'humidity_value must be a' \
            ' number. Got {}'.format(type(data))
        self._humidity_value = data

    @property
    def barometric_pressure(self):
        """Get or set the value of the barometric pressure."""
        return self._barometric_pressure

    @barometric_pressure.setter
    def barometric_pressure(self, data):
        assert isinstance(data, (float, int)), 'barometric_pressure must be a' \
            ' number. Got {}'.format(type(data))
        self._barometric_pressure = data

    @property
    def rain(self):
        """Get or set a boolean for the presence of rain."""
        return self._rain

    @rain.setter
    def rain(self, data):
        self._rain = bool(data)

    @property
    def snow_on_ground(self):
        """Get or set a boolean for the presence of snow on the ground."""
        return self._snow_on_ground

    @snow_on_ground.setter
    def snow_on_ground(self, data):
        self._snow_on_ground = bool(data)

    @property
    def hourly_pressure(self):
        """Get a list of barometric pressures for each hour over the design day."""
        return [self._barometric_pressure] * 24

    def hourly_dew_point_values(self, dry_bulb_condition):
        """Get a list of dew points (C) at each hour over the design day.

        Args:
            dry_bulb_condition: The dry bulb condition for the day.
        """
        hourly_dew_point = []
        max_dpt = self.dew_point(dry_bulb_condition.dry_bulb_max)
        for db in dry_bulb_condition.hourly_values:
            if db >= max_dpt:
                hourly_dew_point.append(max_dpt)
            else:
                hourly_dew_point.append(db)
        return hourly_dew_point

    def dew_point(self, db):
        """Get the dew point (C), which is constant throughout the day (except at saturation).

        Args:
            db: The maximum dry bulb temperature over the day.
        """
        if self._humidity_type == 'Dewpoint':
            return self._humidity_value
        elif self._humidity_type == 'Wetbulb':
            return dew_point_from_db_wb(
                db, self._humidity_value, self._barometric_pressure)
        elif self._humidity_type == 'HumidityRatio':
            return dew_point_from_db_hr(
                db, self._humidity_value, self._barometric_pressure)
        elif self._humidity_type == 'Enthalpy':
            return dew_point_from_db_enth(
                db, self._humidity_value / 1000, self._barometric_pressure)

    def to_dict(self):
        """Convert the Humidity Condition to a dictionary."""
        return {
            'type': 'HumidityCondition',
            'humidity_type': self.humidity_type,
            'humidity_value': self.humidity_value,
            'barometric_pressure': self.barometric_pressure,
            'rain': self.rain,
            'snow_on_ground': self.snow_on_ground
        }

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __copy__(self):
        return HumidityCondition(
            self._humidity_type, self._humidity_value, self._barometric_pressure,
            self._rain, self._snow_on_ground, self.schedule, self.wet_bulb_range)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._humidity_type, self._humidity_value, self._barometric_pressure,
                self._rain, self._snow_on_ground, self.schedule, self.wet_bulb_range)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, HumidityCondition) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Humidity condition representation."""
        return "HumidityCondition [{}: {}]".format(
            self._humidity_type, str(self._humidity_value))


class WindCondition(object):
    """Represents wind conditions on the design day.

    Args:
        wind_speed: Wind speed on the design day [m/s].
        wind_direction: Wind direction on the design day [degrees]. Default: 0

    Properties:
        * wind_speed
        * wind_direction
        * hourly_values
        * hourly_wind_dirs
    """
    __slots__ = ('_wind_speed', '_wind_direction')

    def __init__(self, wind_speed, wind_direction=0):
        """Initialize the class."""
        self.wind_speed = wind_speed
        self.wind_direction = wind_direction

    @classmethod
    def from_dict(cls, data):
        """Create a Wind Condition from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "WindCondition",
            "wind_speed": 3.0,  # float
            "wind_direction": 0.0  # float
            }
        """
        # check required and optional keys
        assert 'wind_speed' in data, 'Required key "wind_speed" is missing!'

        # set defaults for the optional keys
        wind_direction = data['wind_direction'] if 'wind_direction' in data else 0

        return cls(data['wind_speed'], wind_direction)

    @property
    def hourly_values(self):
        """A list of wind speed values for each hour over the design day."""
        return [self._wind_speed] * 24

    @property
    def hourly_wind_dirs(self):
        """A list of wind directions for each hour over the design day."""
        return [self._wind_direction] * 24

    @property
    def wind_speed(self):
        """Get or set the wind speed."""
        return self._wind_speed

    @wind_speed.setter
    def wind_speed(self, data):
        assert isinstance(data, (float, int)), 'wind_speed must be a' \
            ' number. Got {}'.format(type(data))
        self._wind_speed = data

    @property
    def wind_direction(self):
        """Get or set the wind direction."""
        return self._wind_direction

    @wind_direction.setter
    def wind_direction(self, data):
        assert isinstance(data, (float, int)), 'wind_direction must be a' \
            ' number. Got {}'.format(type(data))
        assert 0 <= data <= 360, 'wind_direction {} is not between' \
            ' 0 and 360'.format(data)
        self._wind_direction = data

    def to_dict(self):
        """Convert the Wind Condition to a dictionary."""
        return {
            'type': 'WindCondition',
            'wind_speed': self.wind_speed,
            'wind_direction': self.wind_direction
        }

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __copy__(self):
        return WindCondition(self._wind_speed, self._wind_direction)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._wind_speed, self._wind_direction)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, WindCondition) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Wind condition representation."""
        return "WindCondition [Speed: {}; Dir: {}]".format(
            self.wind_speed, self.wind_direction)


class _SkyCondition(object):
    """An object representing a sky on the design day.

    This is a base class for all solar models and the class of the specific desired
    solar model should typically be used instead of this one.

    Args:
        date: Ladybug Date object for the day of the year on which the design
            day occurs.
        daylight_savings: Boolean to indicate whether daylight savings
            time is active. Default: False
        beam_schedule: Schedule name for beam irradiance. Should be an empty
            string unless the solar model is 'Schedule'.
        diffuse_schedule: Schedule name for diffuse irradiance. Should be an
            empty string unless solar model is 'Schedule'.

    Properties:
        * date
        * daylight_savings
        * beam_schedule
        * diffuse_schedule
        * hourly_sky_cover
    """
    __slots__ = ('_date', '_daylight_savings', 'beam_schedule', 'diffuse_schedule')
    SOLAR_MODELS = ('ASHRAEClearSky', 'ASHRAETau', 'ZhangHuang', 'Schedule')

    def __init__(self, date, daylight_savings=False, beam_schedule='',
                 diffuse_schedule=''):
        self.date = date
        self.daylight_savings = daylight_savings
        self.beam_schedule = beam_schedule
        self.diffuse_schedule = diffuse_schedule

    @classmethod
    def from_dict(cls, data):
        """Create a Sky Condition from a dictionary.

        Args:
            data: A python dictionary in the following format.

        .. code-block:: python

            {
            "type": "SkyCondition",
            "date": [7, 21],
            "daylight_savings": False  # bool
            }
        """
        # process more detailed solar models into their respective classes
        if data['type'] == 'ASHRAEClearSky':
            return ASHRAEClearSky.from_dict(data)
        if data['type'] == 'ASHRAETau':
            return ASHRAETau.from_dict(data)

        # check required keys
        assert 'date' in data, 'Required key "date" is missing!'

        # assign defaults for optional keys
        dl_save = data['daylight_savings'] if 'daylight_savings' \
            in data else False
        beam_schedule = data['beam_schedule'] if 'beam_schedule' in data else ''
        diffuse_schedule = data['diffuse_schedule'] if 'diffuse_schedule' in data else ''

        return cls(Date.from_array(data['date']), dl_save, beam_schedule,
                   diffuse_schedule)

    @property
    def date(self):
        """Get or set a Ladybug Date object for the date of the design day."""
        return self._date

    @date.setter
    def date(self, data):
        assert isinstance(data, Date), 'DesignDay.sky_condition.date must be a' \
            ' Ladybug Date object. Got {}'.format(type(data))
        self._date = data

    @property
    def daylight_savings(self):
        """Get or set a boolean to indicate whether daylight savings time is active."""
        return self._daylight_savings

    @daylight_savings.setter
    def daylight_savings(self, data):
        self._daylight_savings = bool(data)

    @property
    def hourly_sky_cover(self):
        """Get a lists of hourly sky cover values in tenths."""
        return [0] * 24

    def to_dict(self):
        """Convert the Sky Condition to a dictionary."""
        return {
            'type': 'SkyCondition',
            'date': self.date.to_array(),
            'daylight_savings': self.daylight_savings,
            'beam_schedule': self.beam_schedule,
            'diffuse_schedule': self.diffuse_schedule
        }

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def _get_datetimes(self, timestep=1):
        """List of datetimes based on design day date and timestep.

        Note that these datetimes are always in standard time in order to be used
        for solar positions. In other words, they correct from daylight savings time.
        """
        start_moy = self._date.doy * 1440  # convert doy to moy
        if self._daylight_savings:  # spring time forward! (or move the sun back)
            start_moy = start_moy - 60
        if timestep == 1:
            start_moy = start_moy + 30
        num_moys = 24 * timestep
        lp_yr = self._date.leap_year
        return tuple(DateTime.from_moy(start_moy + (i * (1 / timestep) * 60), lp_yr)
                     for i in xrange(num_moys))

    @staticmethod
    def _check_analysis_period(analysis_period):
        """Check an AnalysisPeriod to be sure that it's suitable for a design day."""
        assert isinstance(analysis_period, AnalysisPeriod), 'Expected' \
            ' AnalysisPeriod type. Got {}'.format(type(analysis_period))
        assert analysis_period.st_month == analysis_period.end_month and \
            analysis_period.st_day == analysis_period.end_day and \
            analysis_period.st_hour == 0 and analysis_period.end_hour == 23, \
            'analysis_period "{}" does not represent a single day'.format(
                analysis_period)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __copy__(self):
        return _SkyCondition(self._date, self._daylight_savings, self.beam_schedule,
                             self.diffuse_schedule)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self._date), self._daylight_savings, self.beam_schedule,
                self.diffuse_schedule)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, _SkyCondition) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Sky condition representation."""
        return 'SkyCondition [{}]'.format(self._date)


class ASHRAEClearSky(_SkyCondition):
    """An object representing an original ASHRAE Clear Sky.

    Args:
        date: Ladybug Date object for the day of the year on which the design
            day occurs.
        clearness: Value between 0 and 1.2 that will get multiplied by the model's
            irradiance to correct for factors like elevation.
        daylight_savings: Boolean to indicate whether daylight savings
            time is active. Default: False.

    Properties:
        * date
        * clearness
        * daylight_savings
    """
    __slots__ = ('_clearness',)

    def __init__(self, date, clearness=1, daylight_savings=False):
        _SkyCondition.__init__(self, date, daylight_savings)
        self.clearness = clearness

    @classmethod
    def from_analysis_period(cls, analysis_period, clearness=1, daylight_savings=False):
        """"Initialize a ASHRAEClearSky from an analysis_period"""
        cls._check_analysis_period(analysis_period)
        st_dt = analysis_period.st_time
        return cls(Date(st_dt.month, st_dt.day, st_dt.leap_year), clearness,
                   daylight_savings)

    @classmethod
    def from_dict(cls, data):
        """Create an ASHRAEClearSky condition from a dictionary.

        Args:
            data: A python dictionary in the following format.

        .. code-block:: python

            {
            "type": "ASHRAEClearSky"
            "date": [7, 21],
            "clearness": 0.0  # float,
            "daylight_savings": False  # bool
            }
        """
        # check required and optional keys
        required_keys = ('type', 'date', 'clearness')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        # assign defaults for optional keys
        dl_save = data['daylight_savings'] if 'daylight_savings' \
            in data else False

        return cls(Date.from_array(data['date']), data['clearness'], dl_save)

    @property
    def clearness(self):
        """Get or set the sky clearness."""
        return self._clearness

    @clearness.setter
    def clearness(self, data):
        assert isinstance(data, (float, int)), 'clearness must be a' \
            ' number. Got {}'.format(type(data))
        assert 0 <= data <= 1.2, 'clearness {} is not between 0 and 1.2'.format(data)
        self._clearness = data

    @property
    def hourly_sky_cover(self):
        """Lists of sky cover values in tenths."""
        cover = 0 if self._clearness > 1 else (1 - self._clearness) * 10
        return [cover] * 24

    def radiation_values(self, location, timestep=1):
        """Get arrays of direct, diffuse, and global radiation at each timestep."""
        # create sunpath and get altitude at every timestep of the design day
        sp = Sunpath.from_location(location)
        altitudes = []
        dates = self._get_datetimes(timestep)
        for t_date in dates:
            sun = sp.calculate_sun_from_date_time(t_date)
            altitudes.append(sun.altitude)
        dir_norm, diff_horiz = ashrae_clear_sky(
            altitudes, self._date.month, self._clearness)
        glob_horiz = [dhr + dnr * math.sin(math.radians(alt)) for
                      alt, dnr, dhr in zip(altitudes, dir_norm, diff_horiz)]
        return dir_norm, diff_horiz, glob_horiz

    def to_dict(self):
        """Convert the ASHRAEClearSky to a dictionary."""
        return {
            'type': 'ASHRAEClearSky',
            'date': self.date.to_array(),
            'clearness': self.clearness,
            'daylight_savings': self.daylight_savings,
        }

    def __copy__(self):
        return ASHRAEClearSky(self._date, self._clearness, self._daylight_savings)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self._date), self._clearness, self._daylight_savings)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, ASHRAEClearSky) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """sky condition representation."""
        return 'ASHRAEClearSky [{}]'.format(self._date)


class ASHRAETau(_SkyCondition):
    """An object representing an ASHRAE Revised Clear Sky (Tau model).

    Args:
        date: Ladybug Date object for the day of the year on which the design
            day occurs.
        tau_b: Value for the beam optical depths. Typically found in .stat files.
        tau_d: Value for the diffuse optical depth. Typically found in .stat files.
        use_2017: A boolean to indicate whether the version of the ASHRAE Tau
                model that should be the revised version published in 2017 (True)
                or the original one published in 2009 (False). (Default: False).
        daylight_savings: Boolean to indicate whether daylight savings
            time is active. (Default: False).

    Properties:
        * date
        * tau_b
        * tau_d
        * use_2017
        * daylight_savings
        * hourly_sky_cover
    """
    __slots__ = ('_tau_b', '_tau_d', '_use_2017')

    def __init__(self, date, tau_b, tau_d, use_2017=False, daylight_savings=False):
        _SkyCondition.__init__(self, date, daylight_savings)
        self.tau_b = tau_b
        self.tau_d = tau_d
        self.use_2017 = use_2017

    @classmethod
    def from_analysis_period(cls, analysis_period, tau_b, tau_d, use_2017=False,
                             daylight_savings=False):
        """"Initialize a ASHRAETau sky condition from an analysis_period"""
        cls._check_analysis_period(analysis_period)
        st_dt = analysis_period.st_time
        return cls(Date(st_dt.month, st_dt.day, st_dt.leap_year),
                   tau_b, tau_d, use_2017, daylight_savings)

    @classmethod
    def from_dict(cls, data):
        """Create a ASHRAETau sky condition from a dictionary.

        Args:
            data: A python dictionary in the following format.

        .. code-block:: python

            {
            "type": "ASHRAETau",
            "date": [7, 21],
            "tau_b": 0.0,  # float
            "tau_d": 0.0,  # float
            "use_2017": True,  # boolean for whether the 2017 model is used
            "daylight_savings": False  # bool
            }
        """
        # check required keys
        required_keys = ('date', 'tau_b', 'tau_d')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        # assign defaults for optional keys
        use_2017 = data['use_2017'] if 'use_2017' in data else False
        dl_save = data['daylight_savings'] if 'daylight_savings' \
            in data else False

        return cls(Date.from_array(data['date']),
                   data['tau_b'], data['tau_d'], use_2017, dl_save)

    @property
    def tau_b(self):
        """Get or set the beam optical sky depth."""
        return self._tau_b

    @tau_b.setter
    def tau_b(self, data):
        assert isinstance(data, (float, int)), 'tau_b must be a' \
            ' number. Got {}'.format(type(data))
        self._tau_b = data

    @property
    def tau_d(self):
        """Get or set the diffuse optical sky depth."""
        return self._tau_d

    @tau_d.setter
    def tau_d(self, data):
        assert isinstance(data, (float, int)), 'tau_d must be a' \
            ' number. Got {}'.format(type(data))
        self._tau_d = data

    @property
    def use_2017(self):
        """Boolean for whether the version of the 2017 version of the ASHRAE Tau is used.
        """
        return self._use_2017

    @use_2017.setter
    def use_2017(self, value):
        self._use_2017 = bool(value)

    def radiation_values(self, location, timestep=1):
        """Gat arrays of direct, diffuse, and global radiation at each timestep."""
        # create sunpath and get altitude at every timestep of the design day
        sp = Sunpath.from_location(location)
        altitudes = []
        dates = self._get_datetimes(timestep)
        for t_date in dates:
            sun = sp.calculate_sun_from_date_time(t_date)
            altitudes.append(sun.altitude)
        dir_norm, diff_horiz = ashrae_revised_clear_sky(
            altitudes, self._tau_b, self._tau_d, self._use_2017)
        glob_horiz = [dhr + dnr * math.sin(math.radians(alt)) for
                      alt, dnr, dhr in zip(altitudes, dir_norm, diff_horiz)]
        return dir_norm, diff_horiz, glob_horiz

    def to_dict(self):
        """Convert the Sky Condition to a dictionary."""
        return {
            'type': 'ASHRAETau',
            'date': self.date.to_array(),
            'tau_b': self.tau_b,
            'tau_d': self.tau_d,
            'use_2017': self.use_2017,
            'daylight_savings': self.daylight_savings
        }

    def __copy__(self):
        return ASHRAETau(self._date, self._tau_b, self._tau_d, self._use_2017,
                         self._daylight_savings)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self._date), self._tau_b, self._tau_d, self._use_2017,
                self._daylight_savings)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, ASHRAETau) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """sky condition representation."""
        return 'ASHRAETau [{}]'.format(self._date)



# coding=utf-8
from __future__ import division

import math

from .color import ColorRange
from .datacollection import HourlyContinuousCollection, HourlyDiscontinuousCollection
from .datatype.speed import Speed
from .graphic import GraphicContainer
from .legend import LegendParameters, LegendParametersCategorized
from .compass import Compass

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry2d.mesh import Mesh2D

# Simplify method names
linspace = HourlyContinuousCollection.linspace
histogram = HourlyContinuousCollection.histogram
histogram_circular = HourlyContinuousCollection.histogram_circular


class WindRose(object):
    """Module for calculation and visualization of wind data collection by orientation.

    Args:

        direction_data_collection: A HourlyContinuousCollection or
            HourlyDiscontinuousCollection of wind directions which will be used to
            generate the "bin" intervals for the windrose.
        analysis_data_collection: A HourlyContinuousCollection or
            HourlyDiscontinuousCollection of wind values, corresponding to the wind
            directions, which is "binned" by the calculated direction intervals.
        direction_count: An integer greater than or equal to 3 that determines the
            number of directions to "bin" the wind data by. (Default: 8).

    Properties:
        * direction_data_collection
        * analysis_data_collection
        * direction_values
        * analysis_values
        * analysis_period
        * north
        * histogram_data
        * angles
        * legend
        * legend_parameters
        * base_point
        * compass
        * container
        * prevailing_direction
        * mesh_radius
        * compass_radius
        * colored_mesh
        * color_range
        * orientation_lines
        * frequency_lines
        * windrose_lines
        * frequency_spacing_distance
        * frequency_intervals_compass
        * frequency_intervals_mesh
        * frequency_maximum
        * frequency_hours
    """

    DEFAULT_FREQUENCY_SPACING = 10.0
    DEFAULT_FREQUENCY_HOURS = 200.0
    DEFAULT_BASE_POINT = Point2D()
    DEFAULT_NORTH = 0.0

    def __init__(self, direction_data_collection, analysis_data_collection,
                 direction_count=8):
        """Initialize windrose plot."""
        # Check the input objects
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        assert isinstance(direction_data_collection, acceptable_colls), 'Windrose ' \
            'direction_data_collection must be a HourlyContinuousCollection or ' \
            'HourlyDiscontinuousCollection. Got {}.'.format(
                type(direction_data_collection))
        assert isinstance(analysis_data_collection, acceptable_colls), 'Windrose ' \
            'analysis_data_collection must be a HourlyContinuousCollection or ' \
            'HourlyDiscontinuousCollection. Got {}.'.format(
                type(analysis_data_collection))
        assert direction_data_collection.is_collection_aligned(
            analysis_data_collection), 'Windrose direction_data_collection must be' \
            'aligned with analysis_data_collection. The provided values are not aligned.'
        assert direction_count > 0, 'The number of directions must be ' \
            'greater then one to bin the data, (and greater then three for ' \
            'plotting the wind rose). Currently the direction_count parameter is: '\
            '{}'.format(direction_count)

        # Ensure the analysis period of the data collection has been validated
        if not direction_data_collection.validated_a_period:
            direction_data_collection = \
                direction_data_collection.validate_analysis_period()
        if not analysis_data_collection.validated_a_period:
            analysis_data_collection = \
                analysis_data_collection.validate_analysis_period()

        # Assign the inputs as read-only properties of this data collection
        # get rid of duplicate 0 and 360 values
        direction_data_collection.values = \
            [d % 360.0 for d in direction_data_collection.values]
        self._direction_data_collection = direction_data_collection.to_immutable()
        self._analysis_data_collection = analysis_data_collection.to_immutable()
        self._direction_count = int(direction_count)

        # Compute the windrose data and associated read-only properties
        self._angles = WindRose._compute_angles(direction_count)
        self._is_speed_data_type = isinstance(
            self._analysis_data_collection.header.data_type, Speed)
        self._histogram_data, self._zero_count = \
            self._compute_windrose_data(self.direction_values, self.analysis_values,
                                        self.angles, (0, 360), self._is_speed_data_type)

        # Editable public properties for visualization
        self._legend_parameters = None
        self._frequency_spacing_distance = self.DEFAULT_FREQUENCY_SPACING
        self._frequency_hours = None
        self._frequency_intervals_compass = None
        self._show_freq = True
        self._show_zeros = False
        self._base_point = None
        self._compass = None
        self._north = self.DEFAULT_NORTH
        self._poly_array = None
        self._color_array = None

        # Fixed properties that only need to be computed once
        self._bin_vectors = None
        self._zeros_per_bin = None
        self._real_freq_max = None
        self._prevailing_direction = None
        self._container = None

    @property
    def base_point(self):
        """Get or set a Point2D used as the center point of the windrose geometry.

        (Default: Point3D(0, 0, 0)).
        """
        if self._base_point is None:
            self._base_point = self.DEFAULT_BASE_POINT.duplicate()
        return self._base_point

    @base_point.setter
    def base_point(self, base_point):
        assert isinstance(base_point, type(Point2D())), 'base_point must be a Point2D. '\
            'Got {}.'.format(type(base_point))
        self._compass = None
        self._container = None
        self._base_point = base_point

    @property
    def show_freq(self):
        """Get or set property to control displays of frequency of values per direction.

        (Default: True).
        """
        return self._show_freq

    @show_freq.setter
    def show_freq(self, show_freq):
        self._compass = None
        self._container = None
        self._show_freq = bool(show_freq)

    @property
    def show_zeros(self):
        """Get or set property to control if windrose displays zero values or not.

        (Default: True).
        """
        return self._show_zeros

    @show_zeros.setter
    def show_zeros(self, show_zeros):
        _zero_error = (not self._is_speed_data_type and not show_zeros)
        assert self._is_speed_data_type or _zero_error, 'The calmrose can ' \
            'only be shown if the analysis data type is Speed, for all other data ' \
            'types the zero values will be treated the same as the other numeric ' \
            'values.'
        self._compass = None
        self._container = None
        self._show_zeros = bool(show_zeros)

    @property
    def frequency_spacing_distance(self):
        """Get or set distance for frequency lines (Default: 10 model units).

        To maintain fixed radius, this can be set to: 1.0 / self._legend_segment_count
        """
        return self._frequency_spacing_distance

    @frequency_spacing_distance.setter
    def frequency_spacing_distance(self, val):
        assert val > 0.0 and isinstance(val, (float, int)), \
            'frequency_spacing_distance should be a number greater then 0. ' \
            'Got {}.'.format(val)
        self._compass = None
        self._container = None
        self._frequency_spacing_distance = val

    @property
    def frequency_spacing_hypot_distance(self):
        """Length of radius of windrose assuming perpendicular frequency spacing.

        The limit of the frequency_spacing_hypot_distance as number of directions
        approaches infinity equals the frequency_spacing_distance (as is the case
        with a circle).
        """
        _theta = 360.0 / self._direction_count
        theta = _theta / 2.0 / 180.0 * math.pi
        return self.frequency_spacing_distance / math.cos(theta)

    @property
    def frequency_maximum(self):
        """Get maximum frequency to represent in the windrose plot.

        Default: number of items in largest windrose histogram bin.
        """
        return self.frequency_intervals_compass * self.frequency_hours

    @property
    def frequency_hours(self):
        """Get or set number of hours to bin analysis values, per direction."""
        if self._frequency_hours is None:
            self._frequency_hours = self.DEFAULT_FREQUENCY_HOURS
        return self._frequency_hours

    @frequency_hours.setter
    def frequency_hours(self, frequency_hours):
        assert frequency_hours > 0, 'frequency_hours must be a number greater then 0. ' \
            'Got: {}'.format(frequency_hours)
        self._frequency_hours = int(frequency_hours)

    @property
    def frequency_intervals_compass(self):
        """Get or set the number of intervals in the windrose compass."""
        if self._frequency_intervals_compass is None:
            self._frequency_intervals_compass = self.frequency_intervals_mesh
        return self._frequency_intervals_compass

    @frequency_intervals_compass.setter
    def frequency_intervals_compass(self, frequency_intervals_compass):
        assert frequency_intervals_compass >= 1, 'The frequency_intervals_compass ' \
            'must be greater then 0. Got: {}'.format(frequency_intervals_compass)
        self._compass = None
        self._container = None
        self._frequency_intervals_compass = int(frequency_intervals_compass)

    @property
    def frequency_intervals_mesh(self):
        """Get the number of intervals in the windrose mesh."""
        return int(math.ceil(self.real_freq_max / self.frequency_hours))

    @property
    def legend_parameters(self):
        """Get or set the LegendParameters object for this plot.

        Default segment count is 10.
        """
        return self.container.legend_parameters

    @legend_parameters.setter
    def legend_parameters(self, legend_parameters):
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), 'legend_parameters' \
                ' must be a LegendParameters. Got {}.'.format(type(legend_parameters))
        self._compass = None
        self._container = None
        self._legend_parameters = legend_parameters

    @property
    def north(self):
        """Get or set north orientation for windrose by degrees.

        This must be a number between -360 and 360. 0 is North, 90 is West and 270
        is East. The rotation is applied counterclockwise.
        """
        return self._north

    @north.setter
    def north(self, north):
        assert isinstance(north, (float, int)), 'north must be a number representing ' \
            'degrees. Got {}.'.format(north)

        self._compass = None
        self._container = None
        self._north = float(north) % 360.0

    @property
    def legend(self):
        """Get the Legend object for this plot"""
        return self.container.legend

    @property
    def direction_data_collection(self):
        """Get the direction data collection assigned to this windrose plot."""
        return self._direction_data_collection

    @property
    def analysis_data_collection(self):
        """Get the analysis data collection assigned to this windrose plot."""
        return self._analysis_data_collection

    @property
    def analysis_values(self):
        """Get the hourly wind analysis values to bin by direction (i.e wind speed)."""
        return self._analysis_data_collection.values

    @property
    def direction_values(self):
        """Get the direction data values in this windrose plot."""
        return self._direction_data_collection.values

    @property
    def histogram_data(self):
        """Get a histogram of wind analysis values binned by wind direction.

        This histogram will not contain any zero values, so the total amount of hours
        may not add up to hours in the year. The histogram values per bin will be
        reduced to equal the frequency_maximum property which may occur if the
        frequency_intervals_compass property results in maximum hours less then the
        actual binned data.
        """

        if self._frequency_intervals_compass is not None:
            _real_freq_max = max([len(d) for d in self._histogram_data])
            _freq_int_mesh = int(math.ceil(_real_freq_max / self.frequency_hours))
            if self.frequency_intervals_compass < _freq_int_mesh:
                new_histogram_data = []
                for hbin in self._histogram_data:
                    new_histogram_data.append(hbin[:self.frequency_maximum])
                return tuple(new_histogram_data)

        return self._histogram_data

    @property
    def prevailing_direction(self):
        """Get a tuple of the predominant directions of the wind values.
        """
        if self._prevailing_direction is None:
            dirv_num = self._direction_count
            freqs = [len(b) for b in self._histogram_data]
            dirvs = [i / dirv_num * 360.0 for i in range(dirv_num)]

            # To ensure ties are captured, iterate through check all values.
            max_freq = 0
            max_dirv = []
            for freq, dirv in zip(freqs, dirvs):
                if max_freq < freq:
                    max_freq = freq
                    max_dirv = [dirv]
                elif max_freq == freq:
                    max_dirv.append(dirv)

            self._prevailing_direction = tuple(max_dirv)

        return self._prevailing_direction

    @property
    def zero_count(self):
        """Get the number of wind analysis values equal to zero."""
        return self._zero_count

    @property
    def analysis_period(self):
        """Get the AnalysisPeriod assigned to the hourly plot's data collection."""
        return self._analysis_data_collection.header.analysis_period

    @property
    def angles(self):
        """Get a list of orientation angles used to bin analysis values."""
        return self._angles

    @property
    def bin_vectors(self):
        """Get vectors for orientation intervals."""
        if self._bin_vectors is None:
            self._bin_vectors = WindRose._bin_vectors_radial(self.angles)
        return self._bin_vectors

    @property
    def zeros_per_bin(self):
        """Get the number of analysis values equal to zero, per orientation bin."""
        if self._zeros_per_bin is None:
            self._zeros_per_bin = float(self.zero_count) / (len(self.angles) - 1)
        return self._zeros_per_bin

    @property
    def real_freq_max(self):
        """Get the maximum hours of wind in the largest histogram bin."""
        return max([len(d) for d in self.histogram_data])

    @property
    def compass(self):
        """Get the compass object for the windrose plot.

        Since the setable properties of basepoint, frequency_spacing_distance, and
        legend_parameters all influence the initiation of this object, this property
        is to None if any of those properties are edited by the user.
        """
        if self._compass is None:
            return Compass(self.compass_radius, self.base_point, north_angle=self.north)
        return self._compass

    @property
    def container(self):
        """Get the GraphicContainer for the windrose mesh.

        Since the setable properties of basepoint, frequency_spacing_distance, and
        legend_parameters all influence the initiation of this object, this property
        is to None if any of those properties are edited by the user.
        """
        if self._container is None:
            # Get analysis values
            if self.show_freq:
                values = [b for a in self.histogram_data for b in a]
            else:
                values = [sum(h) / float(len(h)) for h in self.histogram_data]

            if self.show_zeros:
                values += [0 for i in range(self.zero_count)]

            # Create the graphic container
            z = 0
            self._container = GraphicContainer(
                values, self.compass.min_point3d(z), self.compass.max_point3d(z),
                legend_parameters=self._legend_parameters,
                data_type=self.analysis_data_collection.header.data_type,
                unit=self.analysis_data_collection.header.unit)
        return self._container

    @property
    def colored_mesh(self):
        """Get the colored Mesh2D for this graphic.

        Returns:
            A Mesh2D of the wind rose plot.
        """
        assert self._direction_count > 2, 'The number of directions must be ' \
            'greater then three to plot the wind rose. Currently the ' \
            'direction_count parameter is: {}'.format(self._direction_count)
        assert not all(len(h) == 0 for h in self.histogram_data), \
            'No data is available to plot on the wind rose. Mesh cannot be drawn.'

        # Reset computed graphics to account for changes to cached viz properties
        self._compass = None
        self._container = None

        zero_poly_array, zero_color_array = [], []
        max_bar_radius = self.mesh_radius
        min_bar_radius = self._zero_mesh_radius

        if self.show_zeros and self.zero_count > 0:
            # Compute the array for calm rose
            zero_data = [[0] for _ in self.histogram_data]
            zero_data_stacked = [[[0]] for _ in self.histogram_data]
            zero_poly_array, zero_color_array = self._compute_colored_mesh_array(
                zero_data, zero_data_stacked, self.bin_vectors, 0, min_bar_radius,
                show_freq=False)

        # Calculate stacked_data
        flat_data = [b for a in self.histogram_data for b in a]
        max_data = self._legend_parameters.max if self._legend_parameters is not None \
            and self._legend_parameters.max is not None else max(flat_data)
        if self.show_zeros:
            min_data = min(self.analysis_values)
        else:
            min_data = self._legend_parameters.min if self._legend_parameters is not \
                None and self._legend_parameters.min is not None else min(flat_data)
        bin_count = self.legend_parameters.segment_count
        data_range = (min_data, max_data)
        histogram_data_stacked, bin_range = self._histogram_data_nested(
            self.histogram_data, data_range, bin_count)

        # Compute colors
        if not self.show_freq:
            for i in range(self._direction_count):
                stack = histogram_data_stacked[i]
                vals = [b for a in stack for b in a]
                if len(vals) > 0:
                    mean_val = sum(vals) / float(len(vals))
                    histogram_data_stacked[i] = [[mean_val for b in a] for a in stack]
        poly_array, color_array = self._compute_colored_mesh_array(
            self.histogram_data, histogram_data_stacked, self.bin_vectors,
            min_bar_radius, max_bar_radius, self.show_freq)
        poly_array += zero_poly_array
        color_array += zero_color_array

        # Store colors and polygons before processing
        self._poly_array = poly_array
        self._color_array = color_array  # for testing

        # Assign colors
        _color_range = self.color_range
        _color_range = [_color_range.color(val) for val in color_array]
        mesh = Mesh2D.from_face_vertices(poly_array, purge=True)
        mesh.colors = tuple(_color_range)

        # Scale up unit circle to windrose radius (and other transforms)
        return self._transform(mesh)

    @property
    def color_range(self):
        """Get the color range associated with this legend."""
        _l_par = self.legend_parameters

        if isinstance(_l_par, LegendParametersCategorized):
            return ColorRange(_l_par.colors, _l_par.domain, _l_par.continuous_colors)
        else:
            values = self.container.values
            min_val = _l_par.min if _l_par.min is not None else min(values)
            max_val = _l_par.max if _l_par.max is not None else max(values)
            return ColorRange(_l_par.colors, (min_val, max_val))

    @property
    def orientation_lines(self):
        """Get the orientation lines for windrose as a LineSegment2D list."""

        # Reset computed graphics to account for changes to cached viz properties
        self._compass = None
        self._container = None

        # Compute x-axis bin boundaries in polar coordinates
        vec_cpt = (0, 0)
        max_bar_radius = self.compass_radius

        # Vector multiplication with max_bar_radius
        segs = []
        for (vec1, vec2) in self.bin_vectors:
            _seg1 = vec_cpt, (max_bar_radius * vec1[0], max_bar_radius * vec1[1])
            _seg2 = vec_cpt, (max_bar_radius * vec2[0], max_bar_radius * vec2[1])
            segs.extend((LineSegment2D.from_array(_seg1),
                         LineSegment2D.from_array(_seg2)))

        return [self._transform(seg) for seg in segs]

    @property
    def frequency_lines(self):
        """Get the frequency lines for windrose as Polygon2D lists."""

        # Reset computed graphics to account for changes to cached viz properties
        self._compass = None
        self._container = None

        ytick_array = []
        ytick_num = self.frequency_intervals_compass
        zero_dist = self._zero_mesh_radius
        ytick_dist = self.frequency_spacing_hypot_distance
        # Add frequency polygon for calmrose, if exists
        if self.show_zeros and self.zero_count > 0:
            _ytick_array = [(vec1[0] * zero_dist, vec1[1] * zero_dist)
                            for (vec1, _) in self.bin_vectors]
            ytick_array.append(_ytick_array)

        # Add the regular frequency polygons
        for i in range(1, ytick_num + 1):
            _ytick_array = []
            for (vec1, _) in self.bin_vectors:
                _x = (vec1[0] * i * ytick_dist) + (vec1[0] * zero_dist)
                _y = (vec1[1] * i * ytick_dist) + (vec1[1] * zero_dist)
                _ytick_array.append((_x, _y))
            ytick_array.append(_ytick_array)

        return [self._transform(Polygon2D.from_array(vecs)) for vecs in ytick_array]

    @property
    def windrose_lines(self):
        """Get the windrose lines as Polygon2D lists."""

        if self._poly_array is None:
            # colored_mesh property computes windrose geometry with appropriate checks
            # for user-defined parameters, so all windrose geometry properties are set
            # there. If _poly_array is not set, compute colored_mesh.
            self.colored_mesh

        return [self._transform(Polygon2D(vecs)) for vecs in self._poly_array]

    @property
    def mesh_radius(self):
        """Get the radius of the windrose mesh (with zero values).

        This number will not necessarily align to the frequency spacing intervals,
        since fractional values are permitted.
        """
        return self._nonzero_mesh_radius + self._zero_mesh_radius

    @property
    def compass_radius(self):
        """Get the radius of the windrose compass.

        This value is different from the mesh_radius if the frequency_maximum is
        greater then the maximum frequency in the histogram data.
        """
        max_bar_radius = \
            self.frequency_spacing_hypot_distance * self.frequency_intervals_compass

        return max_bar_radius + self._zero_mesh_radius

    @property
    def _nonzero_mesh_radius(self):
        """Get the radius of just the base windrose (excluding the zero values)."""

        ytick_num_frac = self.real_freq_max / self.frequency_hours
        return self.frequency_spacing_hypot_distance * ytick_num_frac

    @property
    def _zero_mesh_radius(self):
        """Get the radius of just the windrose zero values."""

        zero_dist = 0.0
        if self.show_zeros:
            _ytick_num_frac = self.zeros_per_bin / self.frequency_hours
            zero_dist = self.frequency_spacing_hypot_distance * _ytick_num_frac

        return zero_dist

    @staticmethod
    def prevailing_direction_from_data(data, directions_count=8):
        """Get a the prevailing wind direction(s) from a data collection of directions.

        Args:
            data: A data collection with direction angles in them.
            directions_count: An integer for the number of directions around the
                circle to evaluate. (Default: 8).

        Returns:
            A list with at least one value for the prevailing wind direction. This
            will have multiple values in the event of a tie.
        """
        bin_array = WindRose._compute_angles(directions_count)
        hist_data = histogram_circular(data.values, bin_array, (0, 360))
        freqs = [len(b) for b in hist_data]
        dirvs = [i / directions_count * 360.0 for i in range(directions_count)]

        # to ensure ties are captured, iterate through check all values.
        max_freq = 0
        max_dirv = []
        for freq, dirv in zip(freqs, dirvs):
            if max_freq < freq:
                max_freq = freq
                max_dirv = [dirv]
            elif max_freq == freq:
                max_dirv.append(dirv)
        return max_dirv

    @staticmethod
    def _compute_angles(num_of_dir):
        """Compute angles"""

        bin_arr = linspace(0, 360, num_of_dir + 1)

        # Subtract half of segment to capture circular data.
        phi = 360. / (len(bin_arr) - 1) / 2.
        return [b - phi if (b - phi) >= 0.0 else b - phi + 360.
                for b in bin_arr]

    @staticmethod
    def _bin_vectors_radial(bin_arr):
        """Compute the radial coordinates for the histogram bins of values.

        This component will re-orient angles so that 0 degrees starts at the top,
        and increases in a clockwise direction.

        Args:
            bin_arr: List of angles for histogram bar edges in degrees.

        Returns:
            List of vector arrays representing radial histogram.
        """
        vecs = []
        t = 180.0 / math.pi  # for degrees to radian conversion

        # Correction to ensure center of wedge is at the top
        bin_arr = [b - 90. for b in bin_arr]

        # Plot the vectors
        for i in range(len(bin_arr) - 1):
            # Correct range
            theta1 = bin_arr[i]
            theta2 = bin_arr[i + 1]

            # Solve for unit x, y vectors
            # Flip y-component to ensure clockwise rotation of each wedge
            vecs.append(((math.cos(theta1 / t), -math.sin(theta1 / t)),
                         (math.cos(theta2 / t), -math.sin(theta2 / t))))
        return vecs

    @staticmethod
    def _compute_bar_stack_vecs(base_vec_show_freq, vec1, vec2, curr_bar_radius,
                                min_bar_radius, max_bar_num, stacks):
        """Compute the bar geometry arrays for stacked histogram bars.

        Args:
            base_vec_show_freq: list of points for bottem edge of the histogram bar.
            vec1: first bin edge as a vector array.
            vec2: second bin edge as a vector array
            curr_bar_radius: length of the histogram bar.
            min_bar_radius: length from the center of the windrose to the start of the
                histogram bar.
            ytick_curr_num: Number of intervals in the current stacked histogram bar.
            ytick_dist: Length of stack interval.
        Returns:
            List of vector arrays representing a stacked histogram bar.
        """
        bar_interval_vecs = []
        si = -1

        for j, st in enumerate(stacks):
            if len(st) > 0:
                si = j
                break

        if si == -1:
            return bar_interval_vecs

        ytick_dist_inc = min_bar_radius
        for i in range(si, len(stacks)):
            # Stack vectors for interval wedges
            if len(stacks[i]) == 0:
                continue

            ytick_dist_inc += (len(stacks[i]) / max_bar_num * curr_bar_radius)

            # Vector multiplication with y_dist_inc and add to bar_coords
            bar_interval_vecs.append(
                base_vec_show_freq + [
                    (vec1[0] * ytick_dist_inc, vec1[1] * ytick_dist_inc),
                    (vec2[0] * ytick_dist_inc, vec2[1] * ytick_dist_inc)
                ])
            # Update
            base_vec_show_freq = list(reversed(bar_interval_vecs[-1][-2:]))

        return bar_interval_vecs

    @staticmethod
    def _histogram_array_radial(bin_vecs, vec_cpt, hist, hist_stacked, radius_arr,
                                show_freq):
        """Coordinates for a radial histogram as a vector array.

        Args:
            bin_vecs: Array of histogram bin edge vectors.
            vec_cpt: Centerpoint of histogram as tuple.
            hist: Histogram data as a list of lists.
            radius_arr: Minimum and maximum distance for bar radius.
            show_freq: Boolean indicating if stacked histogram.
        Returns:
            List of histogram bars as array of vector coordinates.
        """
        # Get histogram properties for plotting bars
        min_bar_radius, max_bar_radius = radius_arr
        delta_bar_radius = max_bar_radius - min_bar_radius  # does not include zeros
        max_bar_num = max([len(a) for a in hist])  # does not include zeros

        # Plot histogram bar in radial coordinates
        hist_coords = []
        for i, curr_stacks in enumerate(hist_stacked):

            if len([b for a in curr_stacks for b in a]) == 0:
                continue

            # Compute the current bar radius
            (vec2, vec1) = bin_vecs[i]  # flip vecs to be ccw

            # Compute array for bottom of the bar
            if min_bar_radius > 0.0:
                base = [(vec2[0] * min_bar_radius, vec2[1] * min_bar_radius),
                        (vec1[0] * min_bar_radius, vec1[1] * min_bar_radius)]
            else:
                base = [vec_cpt]

            if show_freq:
                bar_vecs = WindRose._compute_bar_stack_vecs(
                    base, vec1, vec2, delta_bar_radius, min_bar_radius,
                    max_bar_num, curr_stacks)
                hist_coords.extend(bar_vecs)
            else:
                curr_bar_num = len([b for a in curr_stacks for b in a])
                frac_len = curr_bar_num / max_bar_num
                curr_bar_radius = delta_bar_radius * frac_len
                bar_vecs = base + [(vec1[0] * (curr_bar_radius + min_bar_radius),
                                    vec1[1] * (curr_bar_radius + min_bar_radius)),
                                   (vec2[0] * (curr_bar_radius + min_bar_radius),
                                    vec2[1] * (curr_bar_radius + min_bar_radius))]
                hist_coords.append(bar_vecs)

        return hist_coords

    @staticmethod
    def _compute_windrose_data(direction_values, analysis_values, bin_array, bin_range,
                               is_speed_data_type):
        """
        Computes the histogram for the windrose.

        Args:
            direction_values: Hourly direction values to bin analysis data by.
            analysis_values: Hourly analysis values.
            bin_intervals: Number of bins for the histogram.
            bin_array: Bin edges as list of direction values.
            bin_range: Maximum and minimum range for histogram.

        Returns:
            The histogram, a proxy histogram to represent zero values, and the
                number of zeros in the analysis values, as a tuple.
        """
        # Filter out zero values if looking at wind speed values
        if is_speed_data_type:
            _direction_values = []
            _analysis_values = []
            for d, v in zip(direction_values, analysis_values):
                if v > 1e-10:
                    _direction_values.append(d)
                    _analysis_values.append(v)
        else:
            _direction_values = direction_values
            _analysis_values = analysis_values

        # Calculate zero rose properties
        zero_count = (len(analysis_values) - len(_analysis_values))

        # Regular hist data
        data = histogram_circular(
            zip(_direction_values, _analysis_values), bin_array, bin_range,
            key=lambda v: v[0])

        # Filter out the direction values and make immutable
        data = tuple(tuple(val[1] for val in bin) for bin in data)

        return data, zero_count

    @staticmethod
    def _histogram_data_nested(histogram_data, analysis_range, bin_count):
        """Computes histogram for each histogram bin based analysis values.

        The function takes a histogram_data structure and applies a second
        pseudo-histogram to each direction bin. In this way the distribution of the
        analysis values are represented. It is a pseudo-histogram since the final
        bin represents values greater than its minimum bounding value, and is therefore
        not bound to a fixed domain.

        Args:
            histogram_data: Histogram of the analysis values binned by direction values,
                as list of analysis arrays.
            analysis_range: Tuple of maximum and minimum analysis values.
            bin_count: The number of bins for the nested pseudo-histogram.

        Returns:
            A tuple with two items.

            - A list of binned analysis values per histogram direction bin.

            - A list of bin intervals used to bin the histogram.
        """
        _histogram_data_nested = []
        min_analy, max_analy = analysis_range[0], analysis_range[1]
        bins = linspace(min_analy, max_analy, int(bin_count))

        # add some epsilon to final bin edge. We do this because the final bin in the
        # analysis histogram is not bounded by a value, but is simply greater then
        # the its minimum bounding value. Note that this means it's not a true histogram
        # with a bounded domain. If this is not done, the final analysis value is not
        # included in the histogram (since the maximum value in a bin interval is
        # exclusive), and would need to be added in it's own bin. There is no way to
        # include the maximum value, without creating arbitrary intervals.
        eps = 1e-1
        bins[-1] += eps

        # Calculate number of hours per frequency segment
        for bar_data_vals in histogram_data:
            _analysis = histogram(bar_data_vals, bins)
            if len(_analysis) == 0:  # all values are the same
                _histogram_data_nested.append([bar_data_vals])
            else:  # it is a normal histogram
                _histogram_data_nested.append(_analysis)

        return _histogram_data_nested, bins

    @staticmethod
    def _compute_colored_mesh_array(hist_data, hist_data_stacked, bin_vecs, min_radius,
                                    max_radius, show_freq):
        """Compute a colored mesh from this object's histogram.

        Args:
            hist_data: Histogram of analysis values greater then zero.
            hist_data_stacked: Histogram of analysis values averaged by number of
                stacks.
            bin_vecs: Array of histogram bin edge vectors.
            min_radius: Minimum radius for windrose mesh.
            max_radius: Maximum radius for windrose mesh.
            show_freq: Boolean indicating if stacked histogram.
        Returns:
            A colored Mesh2D.
        """

        # Default rose is a unit circle centered at origin. We can scale and translate
        # the resulting mesh.
        vec_cpt = (0, 0)

        # Compute histogram array
        hist_coords = WindRose._histogram_array_radial(
            bin_vecs, vec_cpt, hist_data, hist_data_stacked, (min_radius, max_radius),
            show_freq)

        # Construct color array based on show_freq parameter
        color_array = []
        if show_freq:
            for stack in hist_data_stacked:
                for j in range(len(stack)):
                    try:
                        mean_val = sum(stack[j]) / len(stack[j])
                        color_array.append(mean_val)
                    except ZeroDivisionError:
                        mean_val = 0
        else:
            for stack in hist_data_stacked:
                # Value is already averaged
                vals = [b for a in stack for b in a]
                if len(vals) > 0:
                    color_array.append(vals[0])

        # Make mesh
        mesh_array = [[Point2D.from_array(vec) for vec in vecs] for vecs in hist_coords]
        return mesh_array, color_array

    def _transform(self, geometry):
        """Check if geometry defaults and apply transformations."""

        if not self.DEFAULT_BASE_POINT.is_equivalent(self.base_point, 1e-5):
            geometry = geometry.move(self.base_point)
        if abs(self.DEFAULT_NORTH - self.north) > 1e-5:
            geometry = geometry.rotate(self.north / 180.0 * math.pi, self.base_point)

        return geometry

    def __repr__(self):
        """Wind rose Collection representation."""
        return 'Wind Rose:\n\n{}\n\n{}'.format(
            self.direction_data_collection.header,
            self.analysis_data_collection.header)
